"""
Particle Distribution Strategies, how to represent particles in a distribution.
"""

from abc import ABC, abstractmethod
import logging
import numpy as np
from numpy.typing import NDArray

logger = logging.getLogger("particula")


class DistributionStrategy(ABC):
    """
    Abstract base class for particle strategy, defining the common
    interface for mass, radius, and total mass calculations for different
    particle representations.

    Methods:
        get_name: Returns the type of the distribution strategy.
        get_mass: Calculates the mass of particles.
        get_radius: Calculates the radius of particles.
        get_total_mass: Calculates the total mass of particles.
        add_mass: Adds mass to the distribution of particles.
    """

    def get_name(self) -> str:
        """Return the type of the distribution strategy."""
        return self.__class__.__name__

    @abstractmethod
    def get_species_mass(
        self, distribution: NDArray[np.float64], density: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """The mass per species in the particles (or bin).

        Args:
            distribution: The distribution of particle sizes or masses.
            density: The density of the particles.

        Returns:
            NDArray[np.float64]: The mass of the particles
        """

    def get_mass(
        self, distribution: NDArray[np.float64], density: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Calculates the mass of the particles (or bin).

        Args:
            distribution: The distribution of particle sizes or masses.
            density: The density of the particles.

        Returns:
            NDArray[np.float64]: The mass of the particles.
        """
        # one species present
        if distribution.ndim == 1:
            return self.get_species_mass(distribution, density)
        # each column represents a species
        return np.sum(self.get_species_mass(distribution, density), axis=1)

    def get_total_mass(
        self,
        distribution: NDArray[np.float64],
        concentration: NDArray[np.float64],
        density: NDArray[np.float64],
    ) -> np.float64:
        """Calculates the total mass of all particles (or bin).

        Args:
            distribution: The distribution of particle sizes or masses.
            concentration: The concentration of each particle size or mass in
            the distribution.
            density: The density of the particles.

        Returns:
            np.float64: The total mass of the particles.
        """
        # Calculate the mass of each particle and multiply by the concentration
        masses = self.get_mass(distribution, density)
        return np.sum(masses * concentration)

    @abstractmethod
    def get_radius(
        self, distribution: NDArray[np.float64], density: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Calculates the radius of the particles.

        Args:
            distribution: The distribution of particle sizes or masses.
            density: The density of the particles.

        Returns:
            NDArray[np.float64]: The radius of the particles.
        """

    @abstractmethod
    def add_mass(
        self,
        distribution: NDArray[np.float64],
        concentration: NDArray[np.float64],
        density: NDArray[np.float64],
        added_mass: NDArray[np.float64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Adds mass to the distribution of particles.

        Args:
            distribution: The distribution of particle sizes or masses.
            concentration: The concentration of each particle size or mass in
                the distribution.
            density: The density of the particles.
            added_mass: The mass to be added per distribution bin.

        Returns:
            (distribution, concentration): The new distribution array and the
                new concentration array.
        """

    @abstractmethod
    def add_concentration(
        self,
        distribution: NDArray[np.float64],
        concentration: NDArray[np.float64],
        added_distribution: NDArray[np.float64],
        added_concentration: NDArray[np.float64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Adds concentration to the distribution of particles.

        Args:
            distribution: The distribution of particle sizes or masses.
            concentration: The concentration of each particle size or mass in
                the distribution.
            added_distribution: The distribution to be added.
            added_concentration: The concentration to be added.

        Returns:
            (distribution, concentration): The new distribution array and the
                new concentration array.
        """

    @abstractmethod
    def collide_pairs(
        self,
        distribution: NDArray[np.float64],
        concentration: NDArray[np.float64],
        density: NDArray[np.float64],
        indices: NDArray[np.int64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Collides index pairs.

        Args:
            distribution: The distribution of particle sizes or masses.
            concentration: The concentration of each particle size or mass in
                the distribution.
            density: The density of the particles.
            indices: The indices of the particles to collide.

        Returns:
            (distribution, concentration): The new distribution array and the
                new concentration array.
        """


class MassBasedMovingBin(DistributionStrategy):
    """A strategy for particles represented by their mass distribution.

    This strategy calculates particle mass, radius, and total mass based on
    the particle's mass, number concentration, and density. It also moves the
    bins when adding mass to the distribution.
    """

    def get_species_mass(
        self, distribution: NDArray[np.float64], density: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        # In a mass-based strategy, the mass distribution is directly returned.
        return distribution

    def get_radius(
        self, distribution: NDArray[np.float64], density: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        # Calculate the volume of each particle from its mass and density,
        # then calculate the radius.
        volumes = distribution / density
        return (3 * volumes / (4 * np.pi)) ** (1 / 3)

    def add_mass(
        self,
        distribution: NDArray[np.float64],
        concentration: NDArray[np.float64],
        density: NDArray[np.float64],
        added_mass: NDArray[np.float64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        # Add the mass to the distribution moving the bins
        return (distribution + added_mass, concentration)

    def add_concentration(
        self,
        distribution: NDArray[np.float64],
        concentration: NDArray[np.float64],
        added_distribution: NDArray[np.float64],
        added_concentration: NDArray[np.float64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        # check if distribution and added distribution have same elements
        if (
            (distribution.shape != added_distribution.shape) and
            (np.allclose(distribution, added_distribution, rtol=1e-6))
        ):
            message = (
                "When adding concentration to MassBasedMovingBin,"
                + " the distribution and added distribution should have "
                "the same elements. The current distribution shape is "
                f"{distribution.shape} and the added distribution shape is "
                f"{added_distribution.shape}."
            )
            logger.error(message)
            raise ValueError(message)
        # concentration shape should be equal
        if concentration.shape != added_concentration.shape:
            message = (
                "When adding concentration to MassBasedMovingBin,"
                " the concentration and added concentration should have "
                "the same shape. The current concentration shape is "
                f"{concentration.shape} and the added concentration shape is "
                f"{added_concentration.shape}."
            )
            logger.error(message)
            raise ValueError(message)

        # add the concentration in place
        concentration += added_concentration
        return (distribution, concentration)

    def collide_pairs(
        self,
        distribution: NDArray[np.float64],
        concentration: NDArray[np.float64],
        density: NDArray[np.float64],
        indices: NDArray[np.int64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        message = (
            "Colliding pairs in MassBasedMovingBin is not physically"
            + "meaningful, change dyanmic or particle strategy."
        )
        logger.warning(message)
        raise NotImplementedError(message)


class RadiiBasedMovingBin(DistributionStrategy):
    """A strategy for particles represented by their radius.

    This strategy calculates particle mass, radius, and total mass based on
    the particle's radius, number concentration, and density.
    """

    def get_species_mass(
        self, distribution: NDArray[np.float64], density: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        # Calculate the volume of each particle
        volumes = 4 / 3 * np.pi * distribution**3
        return volumes * density  # Mass is volume multiplied by density

    def get_radius(
        self,
        distribution: NDArray[np.float64],
        density: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        return distribution  # Radii are directly available

    def add_mass(
        self,
        distribution: NDArray[np.float64],
        concentration: NDArray[np.float64],
        density: NDArray[np.float64],
        added_mass: NDArray[np.float64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        # Step 1: Calculate mass added per particle
        mass_per_particle = np.where(
            concentration > 0, added_mass / concentration, 0
        )
        # Step 2: Calculate new volumes
        initial_volumes = (4 / 3) * np.pi * np.power(distribution, 3)
        new_volumes = initial_volumes + mass_per_particle / density
        # Step 3: Convert new volumes back to radii
        new_radii = np.power(3 * new_volumes / (4 * np.pi), 1 / 3)
        return (new_radii, concentration)

    def add_concentration(
        self,
        distribution: NDArray[np.float64],
        concentration: NDArray[np.float64],
        added_distribution: NDArray[np.float64],
        added_concentration: NDArray[np.float64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        # check if distribution and added distribution have same elements
        if (distribution.shape != added_distribution.shape) and (
            np.allclose(distribution, added_distribution, rtol=1e-6)
        ):
            message = (
                "When adding concentration to RadiiBasedMovingBin,"
                " the distribution and added distribution should have "
                "the same elements. The current distribution shape is "
                f"{distribution.shape} and the added distribution shape is "
                f"{added_distribution.shape}."
            )
            logger.error(message)
            raise ValueError(message)
        # concentration shape should be equal
        if concentration.shape != added_concentration.shape:
            message = (
                "When adding concentration to RadiiBasedMovingBin,"
                " the concentration and added concentration should have "
                "the same shape. The current concentration shape is "
                f"{concentration.shape} and the added concentration shape is "
                f"{added_concentration.shape}."
            )
            logger.error(message)
            raise ValueError(message)
        # add the concentration in place
        concentration += added_concentration
        return (distribution, concentration)

    def collide_pairs(
        self,
        distribution: NDArray[np.float64],
        concentration: NDArray[np.float64],
        density: NDArray[np.float64],
        indices: NDArray[np.int64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        message = (
            "Colliding pairs in RadiiBasedMovingBin is not physically"
            + "meaningful, change dyanmic or particle strategy."
        )
        logger.warning(message)
        raise NotImplementedError(message)


class SpeciatedMassMovingBin(DistributionStrategy):
    """Strategy for particles with speciated mass distribution.

    Strategy for particles with speciated mass distribution.
    Some particles may have different densities and their mass is
    distributed across different species. This strategy calculates mass,
    radius, and total mass based on the species at each mass, density,
    the particle concentration.
    """

    def get_species_mass(
        self, distribution: NDArray[np.float64], density: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        return distribution

    def get_radius(
        self, distribution: NDArray[np.float64], density: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        # Calculate volume from mass and density, then derive radius
        volumes = np.sum(distribution / density, axis=1)
        return (3 * volumes / (4 * np.pi)) ** (1 / 3)

    def add_mass(
        self,
        distribution: NDArray[np.float64],
        concentration: NDArray[np.float64],
        density: NDArray[np.float64],
        added_mass: NDArray[np.float64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        # Add the mass to the distribution moving the bins
        # limit add to zero, total mass cannot be negative
        # Step 1: Expand concentration if necessary for 2D distributions
        if distribution.ndim == 2:
            concentration_expand = concentration[:, np.newaxis]
        else:
            concentration_expand = concentration
        # Step 2: Calculate mass added per particle (handle zero concentration)
        mass_per_particle = np.where(
            concentration_expand > 0, added_mass / concentration_expand, 0
        )
        # Step 3: Update the distribution by adding the mass per particle
        new_distribution = np.maximum(distribution + mass_per_particle, 0)
        return new_distribution, concentration

    def add_concentration(
        self,
        distribution: NDArray[np.float64],
        concentration: NDArray[np.float64],
        added_distribution: NDArray[np.float64],
        added_concentration: NDArray[np.float64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        # check if distribution and added distribution have same elements
        if (distribution.shape != added_distribution.shape) and (
            np.allclose(distribution, added_distribution, rtol=1e-6)
        ):
            message = (
                "When adding concentration to SpeciatedMassMovingBin,"
                " the distribution and added distribution should have "
                "the same elements. The current distribution shape is "
                f"{distribution.shape} and the added distribution shape is "
                f"{added_distribution.shape}."
            )
            logger.error(message)
            raise ValueError(message)
        # concentration shape should be equal
        if concentration.shape != added_concentration.shape:
            message = (
                "When adding concentration to SpeciatedMassMovingBin,"
                " the concentration and added concentration should have "
                "the same shape. The current concentration shape is "
                f"{concentration.shape} and the added concentration shape is "
                f"{added_concentration.shape}."
            )
            logger.error(message)
            raise ValueError(message)
        # add the concentration in place
        concentration += added_concentration
        return (distribution, concentration)

    def collide_pairs(
        self,
        distribution: NDArray[np.float64],
        concentration: NDArray[np.float64],
        density: NDArray[np.float64],
        indices: NDArray[np.int64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        message = (
            "Colliding pairs in SpeciatedMassMovingBin is not physically"
            + "meaningful, change dyanmic or particle strategy."
        )
        logger.warning(message)
        raise NotImplementedError(message)


class ParticleResolvedSpeciatedMass(DistributionStrategy):
    """Strategy for resolved particles via speciated mass.

    Strategy for resolved particles with speciated mass.
    Particles may have different densities and their mass is
    distributed across different species. This strategy calculates mass,
    radius, and total mass based on the species at each mass, density,
    the particle concentration.
    """

    def get_species_mass(
        self, distribution: NDArray[np.float64], density: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        return distribution

    def get_radius(
        self, distribution: NDArray[np.float64], density: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        # Calculate volume from mass and density, then derive radius
        if distribution.ndim == 1:
            volumes = distribution / density
        else:
            volumes = np.sum(distribution / density, axis=1)
        return (3 * volumes / (4 * np.pi)) ** (1 / 3)

    def add_mass(
        self,
        distribution: NDArray[np.float64],
        concentration: NDArray[np.float64],
        density: NDArray[np.float64],
        added_mass: NDArray[np.float64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        # Add the mass to the distribution moving the bins
        # limit add to zero, total mass cannot be negative
        if distribution.ndim == 2:
            concentration_expand = concentration[:, np.newaxis]
        else:
            concentration_expand = concentration

        new_mass = np.divide(
            np.maximum(distribution * concentration_expand + added_mass, 0),
            concentration_expand,
            out=np.zeros_like(distribution),
            where=concentration_expand != 0
        )
        return (new_mass, concentration)

    def add_concentration(
        self,
        distribution: NDArray[np.float64],
        concentration: NDArray[np.float64],
        added_distribution: NDArray[np.float64],
        added_concentration: NDArray[np.float64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        # Check if added concentration is all ones
        rescaled = False
        if np.all(added_concentration == 1):
            rescaled = True
        if (
            np.allclose(
                added_concentration, np.max(concentration), atol=1e-2) or
            np.all(concentration == 0)
           ):
            # then rescale the added concentration
            added_concentration = added_concentration / np.max(concentration)
            rescaled = True
        if not rescaled:
            message = (
                "When adding concentration to ParticleResolvedSpeciatedMass,"
                + " the added concentration should be all ones or all the same"
                + " value of 1/volume."
            )
            logger.error(message)
            raise ValueError(message)

        # replace concentration with ones, as it is not defined by volume,
        # when it is stored in the class
        concentration = np.divide(
            concentration,
            concentration,
            out=np.zeros_like(concentration),
            where=concentration != 0
        )

        # find empty distribution bins
        empty_bins = np.flatnonzero(np.all(concentration == 0))
        # count of empty bins
        empty_bins_count = len(empty_bins)
        added_bins_count = len(added_concentration)
        if empty_bins_count >= added_bins_count:
            # add all added bins to empty bins
            distribution[empty_bins] = added_distribution
            concentration[empty_bins] = added_concentration
            return distribution, concentration
        # add added bins to empty bins
        if empty_bins_count > 0:
            distribution[empty_bins] = added_distribution[:empty_bins_count]
            concentration[empty_bins] = added_concentration[:empty_bins_count]
        # add the rest of the added bins to the end of the distribution
        distribution = np.concatenate(
            (distribution, added_distribution[empty_bins_count:]),
            axis=0
        )
        concentration = np.concatenate(
            (concentration, added_concentration[empty_bins_count:]),
            axis=0
        )
        return distribution, concentration

    def collide_pairs(
        self,
        distribution: NDArray[np.float64],
        concentration: NDArray[np.float64],
        density: NDArray[np.float64],
        indices: NDArray[np.int64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        # Get the indices of the smaller and larger particles
        small_index = indices[:, 0]
        large_index = indices[:, 1]
        # Check if the distribution is 1D or 2D
        if distribution.ndim == 1:
            # Step 1: Transfer all species mass from smaller to larger
            # particles
            distribution[large_index] += distribution[small_index]
            # Step 2: Zero out the mass of the smaller particles
            distribution[small_index] = 0
            # Step 3: Zero out the concentration of the smaller particles
            concentration[small_index] = 0
            return distribution, concentration

        # Step 1: Transfer all species mass from smaller to larger particles
        distribution[large_index, :] += distribution[small_index, :]
        # Step 2: Zero out the mass of the smaller particles
        distribution[small_index, :] = 0
        # Step 3: Zero out the concentration of the smaller particles
        concentration[small_index] = 0
        # Return the updated distribution (masses) and concentration
        return distribution, concentration



"""
Test of the self_broadcast module in the util package.
"""

import numpy as np

from particula.util.self_broadcast import (
    get_pairwise_sum_matrix,
    get_pairwise_diff_matrix,
    get_pairwise_max_matrix,
)


def test_get_pairwise_sum_matrix():
    """
    Test get_pairwise_sum_matrix with a small array input.
    """
    input_array = np.array([1, 2, 3], dtype=np.float64)

    expected_output = np.array(
        [[2, 3, 4], [3, 4, 5], [4, 5, 6]], dtype=np.float64
    )

    result = get_pairwise_sum_matrix(input_array)

    assert (
        result.shape == expected_output.shape
    ), f"Expected shape {expected_output.shape}, got {result.shape}"
    assert np.allclose(
        result, expected_output
    ), f"Expected {expected_output}, but got {result}"


def test_get_pairwise_diff_matrix():
    """
    Test get_pairwise_diff_matrix with a small array input.
    """
    input_array = np.array([1, 2, 3], dtype=np.float64)

    expected_output = np.array(
        [[0, -1, -2], [1, 0, -1], [2, 1, 0]], dtype=np.float64
    )

    result = get_pairwise_diff_matrix(input_array)

    assert (
        result.shape == expected_output.shape
    ), f"Expected shape {expected_output.shape}, got {result.shape}"
    assert np.allclose(
        result, expected_output
    ), f"Expected {expected_output}, but got {result}"


def test_get_pairwise_max_matrix():
    """
    Test get_pairwise_max_matrix with a small array input.
    """
    input_array = np.array([1, 2, 3], dtype=np.float64)

    expected_output = np.array(
        [[1, 2, 3], [2, 2, 3], [3, 3, 3]], dtype=np.float64
    )

    result = get_pairwise_max_matrix(input_array)

    assert (
        result.shape == expected_output.shape
    ), f"Expected shape {expected_output.shape}, got {result.shape}"
    assert np.allclose(
        result, expected_output
    ), f"Expected {expected_output}, but got {result}"


def test_pairwise_functions_edge_cases():
    """
    Test edge cases including empty arrays and single-element arrays.
    """
    empty_array = np.array([], dtype=np.float64)
    single_element_array = np.array([5], dtype=np.float64)

    # Test empty array
    assert get_pairwise_sum_matrix(empty_array).size == 0
    assert get_pairwise_diff_matrix(empty_array).size == 0
    assert get_pairwise_max_matrix(empty_array).size == 0

    # Test single-element array
    assert np.allclose(
        get_pairwise_sum_matrix(single_element_array), np.array([[10.0]])
    )
    assert np.allclose(
        get_pairwise_diff_matrix(single_element_array), np.array([[0.0]])
    )
    assert np.allclose(
        get_pairwise_max_matrix(single_element_array), np.array([[5.0]])
    )


def test_pairwise_functions_large_array():
    """
    Test that the functions work correctly for a larger array.
    """
    input_array = np.arange(10, dtype=np.float64)  # [0, 1, 2, ..., 9]

    sum_matrix = get_pairwise_sum_matrix(input_array)
    diff_matrix = get_pairwise_diff_matrix(input_array)
    max_matrix = get_pairwise_max_matrix(input_array)

    assert sum_matrix.shape == (10, 10)
    assert diff_matrix.shape == (10, 10)
    assert max_matrix.shape == (10, 10)

    # Check specific expected values
    assert sum_matrix[0, 9] == 9  # 0 + 9
    assert diff_matrix[0, 9] == -9  # 0 - 9
    assert max_matrix[0, 9] == 9  # max(0, 9)
    assert sum_matrix[5, 5] == 10  # 5 + 5
    assert diff_matrix[5, 5] == 0  # 5 - 5
    assert max_matrix[5, 5] == 5  # max(5, 5)



"""Test the gas species builder"""

# pylint: disable=##R0801

import numpy as np
import pytest
from particula.gas.species import GasSpecies
from particula.gas.vapor_pressure_strategies import (
    ConstantVaporPressureStrategy,
)


def test_gas_species_builder_single_species():
    """Test building a single gas species with the GasSpeciesBuilder."""
    vapor_pressure_strategy = ConstantVaporPressureStrategy(
        vapor_pressure=101325
    )
    name = "Oxygen"
    molar_mass = 0.032  # kg/mol
    condensable = False
    concentration = 1.2  # kg/m^3

    gas_species = GasSpecies(
        name=name,
        molar_mass=molar_mass,
        vapor_pressure_strategy=vapor_pressure_strategy,
        condensable=condensable,
        concentration=concentration,
    )

    assert gas_species.name == name
    assert gas_species.get_molar_mass() == molar_mass
    assert gas_species.get_condensable() is condensable
    assert gas_species.get_concentration() == concentration
    assert gas_species.get_pure_vapor_pressure(298) == 101325


def test_gas_species_builder_array_species():
    """Test building an array of gas species with the GasSpeciesBuilder."""
    vapor_pressure_strategy = ConstantVaporPressureStrategy(
        vapor_pressure=np.array([101325, 101325])
    )
    names = np.array(["Oxygen", "Nitrogen"])
    molar_masses = np.array([0.032, 0.028])  # kg/mol
    condensables = np.array([False, False])
    concentrations = np.array([1.2, 0.8])  # kg/m^3

    # Assuming the builder and GasSpecies class can handle array inputs
    # directly.
    gas_species = GasSpecies(
        name=names,
        molar_mass=molar_masses,
        vapor_pressure_strategy=vapor_pressure_strategy,
        condensable=condensables,
        concentration=concentrations,
    )

    assert np.array_equal(gas_species.name, names)
    assert np.array_equal(gas_species.get_molar_mass(), molar_masses)
    assert np.array_equal(gas_species.get_condensable(), condensables)
    assert np.array_equal(gas_species.get_concentration(), concentrations)
    # This assumes get_pure_vapor_pressure can handle and return arrays
    # correctly.
    assert np.array_equal(
        gas_species.get_pure_vapor_pressure(np.array([298, 300])),
        np.array([101325, 101325]),
    )


def test_gas_species_negative_molar_mass():
    """Test gas species with negative molar mass."""
    with pytest.raises(ValueError):
        GasSpecies(
            name="NegativeMass",
            molar_mass=-1,
            condensable=False,
            concentration=1e-10,
        )

    # array input
    with pytest.raises(ValueError):
        GasSpecies(
            name="NegativeMass",
            molar_mass=np.array([-1, 1]),
            condensable=False,
            concentration=np.array([1e-10, 1e-10]),
        )


def test_gas_species_zero_molar_mass():
    """Test gas species with zero molar mass."""
    with pytest.raises(ValueError):
        GasSpecies(
            name="ZeroMass",
            molar_mass=0,
            condensable=False,
            concentration=1e-10,
        )

    # array input
    with pytest.raises(ValueError):
        GasSpecies(
            name="ZeroMass",
            molar_mass=np.array([0, 1]),
            condensable=False,
            concentration=np.array([1e-10, 1e-10]),
        )


def test_gas_species_negative_concentration():
    """Test gas species with negative concentration."""
    with pytest.warns(UserWarning):
        gas_species = GasSpecies(
            name="NegativeConcentration",
            molar_mass=1e10,
            condensable=False,
            concentration=-1,
        )
        assert gas_species.get_concentration() == 0.0

    # array input
    with pytest.warns(UserWarning):
        gas_species = GasSpecies(
            name="NegativeConcentration",
            molar_mass=np.array([1, 1]),
            condensable=False,
            concentration=np.array([-1, 1]),
        )
        assert np.array_equal(gas_species.get_concentration(), [0.0, 1.0])


def test_gas_species_zero_concentration():
    """Test gas species with zero concentration."""
    zero_concentration_species = GasSpecies(
        name="ZeroConcentration",
        molar_mass=1e10,
        condensable=False,
        concentration=0,
    )

    assert zero_concentration_species.get_molar_mass() == 1e10
    assert zero_concentration_species.get_concentration() == 0

    # array input
    zero_concentration_species = GasSpecies(
        name="ZeroConcentration",
        molar_mass=np.array([1, 1]),
        condensable=False,
        concentration=np.array([0, 1]),
    )

    assert np.array_equal(zero_concentration_species.get_molar_mass(), [1, 1])
    assert np.array_equal(
        zero_concentration_species.get_concentration(), [0, 1]
    )


def test_gas_species_get_methods():
    """Test the get methods of GasSpecies."""
    gas_species = GasSpecies(
        name="TestSpecies",
        molar_mass=0.044,
        vapor_pressure_strategy=ConstantVaporPressureStrategy(1000),
        condensable=True,
        concentration=0.5,
    )
    assert gas_species.get_name() == "TestSpecies"
    assert gas_species.get_molar_mass() == 0.044
    assert gas_species.get_condensable() is True
    assert gas_species.get_concentration() == 0.5


def test_gas_species_set_concentration():
    """Test the set_concentration method of GasSpecies."""
    gas_species = GasSpecies(
        name="TestSpecies",
        molar_mass=0.044,
        vapor_pressure_strategy=ConstantVaporPressureStrategy(1000),
        condensable=True,
        concentration=0.5,
    )
    gas_species.set_concentration(1.0)
    assert gas_species.get_concentration() == 1.0
    with pytest.warns(UserWarning):
        gas_species.set_concentration(-1.0)
    assert gas_species.get_concentration() == 0.0


def test_gas_species_add_concentration():
    """Test the add_concentration method of GasSpecies."""
    gas_species = GasSpecies(
        name="TestSpecies",
        molar_mass=0.044,
        vapor_pressure_strategy=ConstantVaporPressureStrategy(1000),
        condensable=True,
        concentration=0.5,
    )
    gas_species.add_concentration(0.5)
    assert gas_species.get_concentration() == 1.0
    with pytest.warns(UserWarning):
        gas_species.add_concentration(-3.0)
    assert gas_species.get_concentration() == 0.0


def test_gas_species_array_operations():
    """Test array operations for concentration in GasSpecies."""
    gas_species = GasSpecies(
        name="TestSpecies",
        molar_mass=0.044,
        vapor_pressure_strategy=ConstantVaporPressureStrategy(1000),
        condensable=True,
        concentration=0.5,
    )
    gas_species.set_concentration(np.array([0.5, 1.0]))
    assert np.array_equal(gas_species.get_concentration(), [0.5, 1.0])
    gas_species.add_concentration(np.array([0.5, -1.0]))
    assert np.array_equal(gas_species.get_concentration(), [1.0, 0.0])



"""Coagulation strategy module.
"""

from abc import ABC, abstractmethod
from typing import Union, Optional
import logging
from numpy.typing import NDArray
import numpy as np

from particula.particles.representation import ParticleRepresentation
from particula.dynamics.coagulation import rate
from particula.dynamics.coagulation.brownian_kernel import (
    brownian_coagulation_kernel_via_system_state,
)
from particula.dynamics.coagulation.particle_resolved_method import (
    particle_resolved_coagulation_step,
)
from particula.dynamics.coagulation.kernel import KernelStrategy
from particula.particles import properties
from particula.gas import properties as gas_properties
from particula.util.reduced_quantity import reduced_self_broadcast

logger = logging.getLogger("particula")


class CoagulationStrategy(ABC):
    """
    Abstract class for defining a coagulation strategy. This class defines the
    methods that must be implemented by any coagulation strategy.

    Methods:
        kernel: Calculate the coagulation kernel.
        loss_rate: Calculate the coagulation loss rate.
        gain_rate: Calculate the coagulation gain rate.
        net_rate: Calculate the net coagulation rate.
        diffusive_knudsen: Calculate the diffusive Knudsen number.
        coulomb_potential_ratio: Calculate the Coulomb potential ratio.
    """

    @abstractmethod
    def dimensionless_kernel(
        self,
        diffusive_knudsen: NDArray[np.float64],
        coulomb_potential_ratio: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """
        Calculate the dimensionless coagulation kernel based on the particle
        properties interactions,
        diffusive Knudsen number and Coulomb potential

        Args:
            diffusive_knudsen: The diffusive Knudsen number
                for the particle [dimensionless].
            coulomb_potential_ratio: The Coulomb potential
                ratio for the particle [dimensionless].

        Returns:
            The dimensionless coagulation kernel for the particle
                [dimensionless].
        """

    @abstractmethod
    def kernel(
        self,
        particle: ParticleRepresentation,
        temperature: float,
        pressure: float,
    ) -> Union[float, NDArray[np.float64]]:
        """
        Calculate the coagulation kernel based on the particle properties,
        temperature, and pressure.

        Args:
            particle: The particle for which the coagulation
                kernel is to be calculated.
            temperature: The temperature of the gas phase [K].
            pressure: The pressure of the gas phase [Pa].

        Returns:
            The coagulation kernel for the particle [m^3/s].
        """

    @abstractmethod
    def loss_rate(
        self,
        particle: ParticleRepresentation,
        kernel: NDArray[np.float64],
    ) -> Union[float, NDArray[np.float64]]:
        """
        Calculate the coagulation loss rate based on the particle radius,
        distribution, and the coagulation kernel.

        Args:
            particle: The particle for which the coagulation
                loss rate is to be calculated.
            kernel: The coagulation kernel.

        Returns:
            The coagulation loss rate for the particle [kg/s].
        """

    @abstractmethod
    def gain_rate(
        self,
        particle: ParticleRepresentation,
        kernel: NDArray[np.float64],
    ) -> Union[float, NDArray[np.float64]]:
        """
        Calculate the coagulation gain rate based on the particle radius,
        distribution, and the coagulation kernel.

        Args:
            particle: The particle for which the coagulation
                gain rate is to be calculated.
            kernel: The coagulation kernel.

        Returns:
            The coagulation gain rate for the particle [kg/s].

        Notes:
            May be abstracted to a separate module when different coagulation
                strategies are implemented (super droplet).
        """

    @abstractmethod
    def net_rate(
        self,
        particle: ParticleRepresentation,
        temperature: float,
        pressure: float,
    ) -> Union[float, NDArray[np.float64]]:
        """
        Calculate the net coagulation rate based on the particle radius,
        distribution, and the coagulation kernel.

        Args:
            particle: The particle class for which the
                coagulation net rate is to be calculated.
            temperature: The temperature of the gas phase [K].
            pressure: The pressure of the gas phase [Pa].

        Returns:
            Union[float, NDArray[np.float64]]: The net coagulation rate for the
                particle [kg/s].
        """

    @abstractmethod
    def step(
        self,
        particle: ParticleRepresentation,
        temperature: float,
        pressure: float,
        time_step: float,
    ) -> ParticleRepresentation:
        """
        Perform a single step of the coagulation process.

        Args:
            particle: The particle for which the coagulation step
                is to be performed.
            temperature: The temperature of the gas phase [K].
            pressure: The pressure of the gas phase [Pa].
            time_step: The time step for the coagulation process [s].

        Returns:
            ParticleRepresentation: The particle after the coagulation step.
        """

    def diffusive_knudsen(
        self,
        particle: ParticleRepresentation,
        temperature: float,
        pressure: float,
    ) -> NDArray[np.float64]:
        """
        Calculate the diffusive Knudsen number based on the particle
        properties, temperature, and pressure.

        Args:
            particle: The particle for which the diffusive
                Knudsen number is to be calculated.
            temperature: The temperature of the gas phase [K].
            pressure: The pressure of the gas phase [Pa].

        Returns:
            NDArray[np.float64]: The diffusive Knudsen number for the particle
                [dimensionless].
        """
        # properties calculation for friction factor
        friction_factor = self.friction_factor(
            particle=particle, temperature=temperature, pressure=pressure
        )
        # coulomb potential ratio
        coulomb_potential_ratio = self.coulomb_potential_ratio(
            particle=particle, temperature=temperature
        )
        return properties.diffusive_knudsen_number(
            radius=particle.get_radius(),
            mass_particle=particle.get_mass(),
            friction_factor=friction_factor,
            coulomb_potential_ratio=coulomb_potential_ratio,
            temperature=temperature,
        )  # type: ignore

    def coulomb_potential_ratio(
        self, particle: ParticleRepresentation, temperature: float
    ) -> NDArray[np.float64]:
        """
        Calculate the Coulomb potential ratio based on the particle properties
        and temperature.

        Args:
            particle: The particles for which the Coulomb
                potential ratio is to be calculated.
            temperature: The temperature of the gas phase [K].

        Returns:
            The Coulomb potential ratio for the particle
                [dimensionless].
        """
        return properties.coulomb_enhancement.ratio(
            radius=particle.get_radius(),
            charge=particle.get_charge(),
            temperature=temperature,
        )  # type: ignore

    def friction_factor(
        self,
        particle: ParticleRepresentation,
        temperature: float,
        pressure: float,
    ) -> NDArray[np.float64]:
        """
        Calculate the friction factor based on the particle properties,
        temperature, and pressure.

        Args:
            particle: The particle for which the friction factor
                is to be calculated.
            temperature: The temperature of the gas phase [K].
            pressure: The pressure of the gas phase [Pa].

        Returns:
            The friction factor for the particle [dimensionless].
        """
        dynamic_viscosity = gas_properties.get_dynamic_viscosity(
            temperature=temperature  # assume standard atmospheric composition
        )
        mean_free_path = gas_properties.molecule_mean_free_path(
            temperature=temperature,
            pressure=pressure,
            dynamic_viscosity=dynamic_viscosity,
        )
        knudsen_number = properties.calculate_knudsen_number(
            mean_free_path=mean_free_path,
            particle_radius=particle.get_radius(),
        )
        slip_correction = properties.cunningham_slip_correction(
            knudsen_number=knudsen_number
        )
        return properties.friction_factor(
            radius=particle.get_radius(),
            dynamic_viscosity=dynamic_viscosity,
            slip_correction=slip_correction,
        )  # type: ignore


# Define a coagulation strategy
class DiscreteSimple(CoagulationStrategy):
    """
    Discrete Brownian coagulation strategy class. This class implements the
    methods defined in the CoagulationStrategy abstract class.

    Methods:
    --------
    - kernel: Calculate the coagulation kernel.
    - loss_rate: Calculate the coagulation loss rate.
    - gain_rate: Calculate the coagulation gain rate.
    - net_rate: Calculate the net coagulation rate.
    """

    def dimensionless_kernel(
        self,
        diffusive_knudsen: NDArray[np.float64],
        coulomb_potential_ratio: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        message = (
            "Dimensionless kernel not implemented in simple "
            + "coagulation strategy."
        )
        logger.error(message)
        raise NotImplementedError(message)

    def kernel(
        self,
        particle: ParticleRepresentation,
        temperature: float,
        pressure: float,
    ) -> Union[float, NDArray[np.float64]]:

        return brownian_coagulation_kernel_via_system_state(
            radius_particle=particle.get_radius(),
            mass_particle=particle.get_mass(),
            temperature=temperature,
            pressure=pressure,
        )

    def loss_rate(
        self,
        particle: ParticleRepresentation,
        kernel: NDArray[np.float64],
    ) -> Union[float, NDArray[np.float64]]:

        return rate.discrete_loss(
            concentration=particle.concentration,
            kernel=kernel,
        )

    def gain_rate(
        self,
        particle: ParticleRepresentation,
        kernel: NDArray[np.float64],
    ) -> Union[float, NDArray[np.float64]]:

        return rate.discrete_gain(
            radius=particle.get_radius(),
            concentration=particle.concentration,
            kernel=kernel,
        )

    def net_rate(
        self,
        particle: ParticleRepresentation,
        temperature: float,
        pressure: float,
    ) -> Union[float, NDArray[np.float64]]:

        kernel = self.kernel(
            particle=particle, temperature=temperature, pressure=pressure
        )
        loss_rate = self.loss_rate(
            particle=particle, kernel=kernel  # type: ignore
        )
        gain_rate = self.gain_rate(
            particle=particle, kernel=kernel  # type: ignore
        )
        return gain_rate - loss_rate

    def step(
        self,
        particle: ParticleRepresentation,
        temperature: float,
        pressure: float,
        time_step: float,
    ) -> ParticleRepresentation:

        particle.add_concentration(
            self.net_rate(  # type: ignore
                particle=particle, temperature=temperature, pressure=pressure
            )
            * time_step
        )
        return particle


class DiscreteGeneral(CoagulationStrategy):
    """
    Discrete general coagulation strategy class. This class implements the
    methods defined in the CoagulationStrategy abstract class. The kernel
    strategy is passed as an argument to the class, to use a dimensionless
    kernel representation.

    Attributes:
    -----------
    - kernel_strategy: The kernel strategy to be used for the coagulation, from
    the KernelStrategy class.

    Methods:
    --------
    - kernel: Calculate the coagulation kernel.
    - loss_rate: Calculate the coagulation loss rate.
    - gain_rate: Calculate the coagulation gain rate.
    - net_rate: Calculate the net coagulation rate.
    """

    def __init__(self, kernel_strategy: KernelStrategy):
        self.kernel_strategy = kernel_strategy

    def dimensionless_kernel(
        self,
        diffusive_knudsen: NDArray[np.float64],
        coulomb_potential_ratio: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        return self.kernel_strategy.dimensionless(
            diffusive_knudsen=diffusive_knudsen,
            coulomb_potential_ratio=coulomb_potential_ratio,
        )

    def kernel(
        self,
        particle: ParticleRepresentation,
        temperature: float,
        pressure: float,
    ) -> Union[float, NDArray[np.float64]]:
        diffusive_knudsen = self.diffusive_knudsen(
            particle=particle, temperature=temperature, pressure=pressure
        )
        coulomb_potential_ratio = self.coulomb_potential_ratio(
            particle=particle, temperature=temperature
        )
        dimensionless_kernel = self.dimensionless_kernel(
            diffusive_knudsen=diffusive_knudsen,
            coulomb_potential_ratio=coulomb_potential_ratio,
        )
        friction_factor = self.friction_factor(
            particle=particle, temperature=temperature, pressure=pressure
        )
        # Calculate the pairwise sum of radii
        radius = particle.get_radius()
        sum_of_radii = radius[:, np.newaxis] + radius
        # square matrix of mass
        reduced_mass = reduced_self_broadcast(particle.get_mass())
        # square matrix of friction factor
        reduced_friction_factor = reduced_self_broadcast(friction_factor)

        return self.kernel_strategy.kernel(
            dimensionless_kernel=dimensionless_kernel,
            coulomb_potential_ratio=coulomb_potential_ratio,
            sum_of_radii=sum_of_radii,
            reduced_mass=reduced_mass,
            reduced_friction_factor=reduced_friction_factor,
        )

    def loss_rate(
        self,
        particle: ParticleRepresentation,
        kernel: NDArray[np.float64],
    ) -> Union[float, NDArray[np.float64]]:

        return rate.discrete_loss(
            concentration=particle.concentration,
            kernel=kernel,
        )

    def gain_rate(
        self,
        particle: ParticleRepresentation,
        kernel: NDArray[np.float64],
    ) -> Union[float, NDArray[np.float64]]:

        return rate.discrete_gain(
            radius=particle.get_radius(),
            concentration=particle.concentration,
            kernel=kernel,
        )

    def net_rate(
        self,
        particle: ParticleRepresentation,
        temperature: float,
        pressure: float,
    ) -> Union[float, NDArray[np.float64]]:

        kernel = self.kernel(
            particle=particle, temperature=temperature, pressure=pressure
        )
        loss_rate = self.loss_rate(
            particle=particle, kernel=kernel  # type: ignore
        )
        gain_rate = self.gain_rate(
            particle=particle, kernel=kernel  # type: ignore
        )
        return gain_rate - loss_rate

    def step(
        self,
        particle: ParticleRepresentation,
        temperature: float,
        pressure: float,
        time_step: float,
    ) -> ParticleRepresentation:

        particle.add_concentration(
            self.net_rate(  # type: ignore
                particle=particle, temperature=temperature, pressure=pressure
            )
            * time_step
        )
        return particle


class ContinuousGeneralPDF(CoagulationStrategy):
    """
    Continuous PDF coagulation strategy class. This class implements the
    methods defined in the CoagulationStrategy abstract class. The kernel
    strategy is passed as an argument to the class, should use a dimensionless
    kernel representation.

    Methods:
        kernel: Calculate the coagulation kernel.
        loss_rate: Calculate the coagulation loss rate.
        gain_rate: Calculate the coagulation gain rate.
        net_rate: Calculate the net coagulation rate.
    """

    def __init__(self, kernel_strategy: KernelStrategy):
        self.kernel_strategy = kernel_strategy

    def dimensionless_kernel(
        self,
        diffusive_knudsen: NDArray[np.float64],
        coulomb_potential_ratio: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        return self.kernel_strategy.dimensionless(
            diffusive_knudsen=diffusive_knudsen,
            coulomb_potential_ratio=coulomb_potential_ratio,
        )

    def kernel(
        self,
        particle: ParticleRepresentation,
        temperature: float,
        pressure: float,
    ) -> Union[float, NDArray[np.float64]]:
        diffusive_knudsen = self.diffusive_knudsen(
            particle=particle, temperature=temperature, pressure=pressure
        )
        coulomb_potential_ratio = self.coulomb_potential_ratio(
            particle=particle, temperature=temperature
        )
        dimensionless_kernel = self.dimensionless_kernel(
            diffusive_knudsen=diffusive_knudsen,
            coulomb_potential_ratio=coulomb_potential_ratio,
        )
        friction_factor = self.friction_factor(
            particle=particle, temperature=temperature, pressure=pressure
        )
        # Calculate the pairwise sum of radii
        radius = particle.get_radius()
        sum_of_radii = radius[:, np.newaxis] + radius
        # square matrix of mass
        reduced_mass = reduced_self_broadcast(particle.get_mass())
        # square matrix of friction factor
        reduced_friction_factor = reduced_self_broadcast(friction_factor)

        return self.kernel_strategy.kernel(
            dimensionless_kernel=dimensionless_kernel,
            coulomb_potential_ratio=coulomb_potential_ratio,
            sum_of_radii=sum_of_radii,
            reduced_mass=reduced_mass,
            reduced_friction_factor=reduced_friction_factor,
        )

    def loss_rate(
        self,
        particle: ParticleRepresentation,
        kernel: NDArray[np.float64],
    ) -> Union[float, NDArray[np.float64]]:

        return rate.continuous_loss(
            radius=particle.get_radius(),
            concentration=particle.concentration,
            kernel=kernel,
        )

    def gain_rate(
        self,
        particle: ParticleRepresentation,
        kernel: NDArray[np.float64],
    ) -> Union[float, NDArray[np.float64]]:

        return rate.continuous_gain(
            radius=particle.get_radius(),
            concentration=particle.concentration,
            kernel=kernel,
        )

    def net_rate(
        self,
        particle: ParticleRepresentation,
        temperature: float,
        pressure: float,
    ) -> Union[float, NDArray[np.float64]]:

        kernel = self.kernel(
            particle=particle, temperature=temperature, pressure=pressure
        )
        loss_rate = self.loss_rate(
            particle=particle, kernel=kernel  # type: ignore
        )
        gain_rate = self.gain_rate(
            particle=particle, kernel=kernel  # type: ignore
        )
        return gain_rate - loss_rate

    def step(
        self,
        particle: ParticleRepresentation,
        temperature: float,
        pressure: float,
        time_step: float,
    ) -> ParticleRepresentation:

        particle.add_concentration(
            self.net_rate(  # type: ignore
                particle=particle, temperature=temperature, pressure=pressure
            )
            * time_step
        )
        return particle


class ParticleResolved(CoagulationStrategy):
    """
    Particle-resolved coagulation strategy class. This class implements the
    methods defined in the CoagulationStrategy abstract class. The kernel
    strategy is passed as an argument to the class, should use a dimensionless
    kernel representation.

    Methods:
        kernel: Calculate the coagulation kernel.
        loss_rate: Not implemented.
        gain_rate: Not implemented.
        net_rate: Not implemented.
        step: Perform a single step of the coagulation process.
    """

    def __init__(
        self,
        kernel_radius: Optional[NDArray[np.float64]] = None,
        kernel_bins_number: Optional[int] = None,
        kernel_bins_per_decade: int = 10,
    ):
        self.kernel_radius = kernel_radius
        self.kernel_bins_number = kernel_bins_number
        self.kernel_bins_per_decade = kernel_bins_per_decade

    def get_kernel_radius(
        self, particle: ParticleRepresentation
    ) -> NDArray[np.float64]:
        """Get the binning for the kernel radius.

        If the kernel radius is not set, it will be calculated based on the
        particle radius.

        Args:
            particle: The particle for which the kernel radius is to be
                calculated.

        Returns:
            The kernel radius for the particle [m].
        """
        if self.kernel_radius is not None:
            return self.kernel_radius
        # else find the non-zero min and max radii, the log space them
        min_radius = np.min(particle.get_radius()[particle.get_radius() > 0])
        max_radius = np.max(particle.get_radius()[particle.get_radius() > 0])
        if self.kernel_bins_number is not None:
            return np.logspace(
                np.log10(min_radius),
                np.log10(max_radius),
                num=self.kernel_bins_number,
                base=10,
                dtype=np.float64,
            )
        # else kernel bins per decade
        num = np.ceil(
            self.kernel_bins_per_decade * np.log10(max_radius / min_radius),
        )
        return np.logspace(
            np.log10(min_radius),
            np.log10(max_radius),
            num=int(num),
            base=10,
            dtype=np.float64,
        )

    def dimensionless_kernel(
        self,
        diffusive_knudsen: NDArray[np.float64],
        coulomb_potential_ratio: NDArray[np.float64],
    ) -> NDArray[np.float64]:

        message = (
            "Dimensionless kernel not implemented in simple "
            + "coagulation strategy."
        )
        logger.error(message)
        raise NotImplementedError(message)

    def kernel(
        self,
        particle: ParticleRepresentation,
        temperature: float,
        pressure: float,
    ) -> Union[float, NDArray[np.float64]]:
        # need to update later with the correct mass dependency
        radius_bins = self.get_kernel_radius(particle)
        mass_bins = (
            4 / 3 * np.pi * np.power(radius_bins, 3) * 1000  # type: ignore
        )
        return brownian_coagulation_kernel_via_system_state(
            radius_particle=radius_bins,  # type: ignore
            mass_particle=mass_bins,  # type: ignore
            temperature=temperature,
            pressure=pressure,
        )

    def loss_rate(
        self,
        particle: ParticleRepresentation,
        kernel: NDArray[np.float64],
    ) -> Union[float, NDArray[np.float64]]:

        message = "Loss rate not implemented in particle-resolved coagulation."
        logger.error(message)
        raise NotImplementedError(message)

    def gain_rate(
        self,
        particle: ParticleRepresentation,
        kernel: NDArray[np.float64],
    ) -> Union[float, NDArray[np.float64]]:

        message = "Gain rate not implemented in particle-resolved coagulation."
        logger.error(message)
        raise NotImplementedError(message)

    def net_rate(
        self,
        particle: ParticleRepresentation,
        temperature: float,
        pressure: float,
    ) -> Union[float, NDArray[np.float64]]:

        message = "Net rate not implemented in particle-resolved coagulation."
        logger.error(message)
        raise NotImplementedError(message)

    def step(
        self,
        particle: ParticleRepresentation,
        temperature: float,
        pressure: float,
        time_step: float,
    ) -> ParticleRepresentation:

        # need to add the particle resolved coagulation step
        loss_gain_indices = particle_resolved_coagulation_step(
            particle_radius=particle.get_radius(),
            kernel=self.kernel(  # type: ignore
                particle=particle, temperature=temperature, pressure=pressure
            ),
            kernel_radius=self.get_kernel_radius(particle),
            volume=particle.volume,
            time_step=time_step,
            random_generator=np.random.default_rng(),
        )
        particle.collide_pairs(loss_gain_indices)
        return particle



"""
Test Special Functions
"""

import numpy as np
from particula.particles.properties.special_functions import (
    debye_function,
)


def test_debye_function_single_float():
    """Test the Debye function for single float values, including extremes."""
    assert np.isclose(debye_function(1.0), 0.7765038970390566)
    assert np.isclose(debye_function(1.0, n=2), 0.7078773477535959)
    assert np.isclose(
        debye_function(1e-10, integration_points=10000), 1.0, atol=1e-4
    )
    assert np.isclose(debye_function(1e10), 0.0)
    assert np.isclose(debye_function(1e-10, n=2), 1.0)
    assert np.isclose(debye_function(1e10, n=2), 0.0)


def test_debye_function_numpy_array():
    """Test the Debye function for a numpy array."""
    input_array = np.array([1.0, 2.0, 3.0])
    expected_output = np.array([0.7765039, 0.60594683, 0.47943507])
    result = debye_function(input_array)
    assert np.allclose(result, expected_output)


def test_debye_function_with_integration_points():
    """Test Debye function with lower integration points."""
    result = debye_function(1.0, integration_points=100)
    assert np.isclose(result, 0.7674304601560753)

    result = debye_function(1.0, integration_points=100, n=2)
    assert np.isclose(result, 0.7077640915847944)



""" Module for calculating the dynamic viscosity

    The dynamic viscosity is calculated using the Sutherland formula,
    assuming ideal gas behavior, as a function of temperature.

    "The dynamic viscosity equals the product of the sum of
    Sutherland's constant and the reference temperature divided by
    the sum of Sutherland's constant and the temperature,
    the reference viscosity and the ratio to the 3/2 power
    of the temperature to reference temperature."

    https://resources.wolframcloud.com/FormulaRepository/resources/Sutherlands-Formula
"""

import logging

from particula.util.constants import (REF_TEMPERATURE_STP,
                                      REF_VISCOSITY_AIR_STP,
                                      SUTHERLAND_CONSTANT)

logger = logging.getLogger("particula")  # get instance of logger


def get_dynamic_viscosity(
    temperature: float,
    reference_viscosity: float = REF_VISCOSITY_AIR_STP,
    reference_temperature: float = REF_TEMPERATURE_STP
) -> float:
    """
    Calculates the dynamic viscosity of air via Sutherland's formula, which is
    a common method in fluid dynamics for gases that involves temperature
    adjustments.

    Args:
    -----
    - temperature: Desired air temperature [K]. Must be greater than 0.
    - reference_viscosity: Gas viscosity [Pa*s] at the reference temperature
    (default is STP).
    - reference_temperature: Gas temperature [K] for the reference viscosity
    (default is STP).

    Returns:
    --------
    - float: The dynamic viscosity of air at the given temperature [Pa*s].

    Raises:
    ------
    - ValueError: If the temperature is less than or equal to 0.

    References:
    ----------
    https://resources.wolframcloud.com/FormulaRepository/resources/Sutherlands-Formula
    """
    if temperature <= 0:
        logger.error("Temperature must be greater than 0 Kelvin.")
        raise ValueError("Temperature must be greater than 0 Kelvin.")
    return (
        reference_viscosity * (temperature / reference_temperature)**1.5 *
        (reference_temperature + SUTHERLAND_CONSTANT) /
        (temperature + SUTHERLAND_CONSTANT)
    )



"""Particle settling velocity in a fluid."""

from typing import Union
import numpy as np
from numpy.typing import NDArray

from particula.util.constants import STANDARD_GRAVITY
from particula.util.validate_inputs import validate_inputs
from particula.gas.properties.dynamic_viscosity import (
    get_dynamic_viscosity,
)
from particula.gas.properties.mean_free_path import (
    molecule_mean_free_path,
)
from particula.particles.properties.slip_correction_module import (
    cunningham_slip_correction,
)
from particula.particles.properties.knudsen_number_module import (
    calculate_knudsen_number,
)


def particle_settling_velocity(
    particle_radius: Union[float, NDArray[np.float64]],
    particle_density: Union[float, NDArray[np.float64]],
    slip_correction_factor: Union[float, NDArray[np.float64]],
    dynamic_viscosity: float,
    gravitational_acceleration: float = STANDARD_GRAVITY,
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the settling velocity of a particle in a fluid.

    Arguments:
        particle_radius: The radius of the particle [m].
        particle_density: The density of the particle [kg/m³].
        slip_correction_factor: The slip correction factor to
            account for non-continuum effects [dimensionless].
        gravitational_acceleration: The gravitational acceleration.
            Defaults to standard gravity [9.80665 m/s²].
        dynamic_viscosity: The dynamic viscosity of the fluid [Pa*s].

    Returns:
        The settling velocity of the particle in the fluid [m/s].

    """

    # Calculate the settling velocity using the given formula
    settling_velocity = (
        (2 * particle_radius) ** 2
        * particle_density
        * slip_correction_factor
        * gravitational_acceleration
        / (18 * dynamic_viscosity)
    )

    return settling_velocity


@validate_inputs(
    {
        "particle_inertia_time": "positive",
        "gravitational_acceleration": "positive",
        "slip_correction_factor": "positive",
    }
)
def get_particle_settling_velocity_via_inertia(
    particle_inertia_time: Union[float, NDArray[np.float64]],
    gravitational_acceleration: float,
    slip_correction_factor: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the gravitational settling velocity from the particle inertia
    time.

    The settling velocity (v_s) is given by:

        v_s = g * τ_p * C_c

    - v_s : Settling velocity [m/s]
    - g (gravitational_acceleration) : Gravitational acceleration [m/s²]
    - τ_p (particle_inertia_time) : Particle inertia time [s]
    - C_c (slip_correction_factor) : Cunningham slip correction factor [-]

    Arguments:
    ----------
        - particle_inertia_time : Particle inertia time [s]
        - gravitational_acceleration : Gravitational acceleration [m/s²]
        - slip_correction_factor : Cunningham slip correction factor [-]

    Returns:
    --------
        - Particle settling velocity [m/s]
    """
    return (
        gravitational_acceleration
        * particle_inertia_time
        * slip_correction_factor
    )


def particle_settling_velocity_via_system_state(
    particle_radius: Union[float, NDArray[np.float64]],
    particle_density: Union[float, NDArray[np.float64]],
    temperature: float,
    pressure: float,
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the settling velocity of a particle.

    Arguments:
        particle_radius: The radius of the particle in meters (m).
        particle_density: The density of the particle (kg/m³).
        temperature: The temperature of the system in Kelvin (K).
        pressure: The pressure of the system in Pascals (Pa).

    Returns:
        The settling velocity of the particle in meters per second (m/s).
    """

    # Step 1: Calculate the dynamic viscosity of the gas
    dynamic_viscosity = get_dynamic_viscosity(temperature=temperature)

    # Step 2: Calculate the mean free path of the gas molecules
    mean_free_path = molecule_mean_free_path(
        temperature=temperature,
        pressure=pressure,
        dynamic_viscosity=dynamic_viscosity,
    )

    # Step 3: Calculate the Knudsen number (characterizes flow regime)
    knudsen_number = calculate_knudsen_number(
        mean_free_path=mean_free_path, particle_radius=particle_radius
    )

    # Step 4: Calculate the slip correction factor (Cunningham correction)
    slip_correction_factor = cunningham_slip_correction(
        knudsen_number=knudsen_number,
    )

    # Step 5: Calculate the particle settling velocity
    return particle_settling_velocity(
        particle_radius=particle_radius,
        particle_density=particle_density,
        slip_correction_factor=slip_correction_factor,
        dynamic_viscosity=dynamic_viscosity,
    )



"""This module contains the GasSpeciesBuilder class, which is a builder class
for GasSpecies objects. The GasSpeciesBuilder class allows for a more fluent
and readable creation of GasSpecies as this class provides validation and
unit conversion for the parameters of the GasSpecies object.
"""

from typing import Union
import logging
from numpy.typing import NDArray
import numpy as np
from particula.abc_builder import (
    BuilderABC,
)
from particula.builder_mixin import (
    BuilderMolarMassMixin,
    BuilderConcentrationMixin,
)
from particula.gas.species import GasSpecies
from particula.gas.vapor_pressure_strategies import (
    VaporPressureStrategy,
    ConstantVaporPressureStrategy,
)

logger = logging.getLogger("particula")


class GasSpeciesBuilder(
    BuilderABC, BuilderMolarMassMixin, BuilderConcentrationMixin
):
    """Builder class for GasSpecies objects, allowing for a more fluent and
    readable creation of GasSpecies instances with optional parameters.

    Attributes:
        - name : The name of the gas species.
        - molar_mass : The molar mass of the gas species in kg/mol.
        - vapor_pressure_strategy : The vapor pressure strategy for the
            gas species.
        - condensable : Whether the gas species is condensable.
        - concentration : The concentration of the gas species in the
            mixture, in kg/m^3.

    Raises:
        - ValueError : If any required key is missing. During check_keys and
            pre_build_check. Or if trying to set an invalid parameter.
        - Warning : If using default units for any parameter.
    """

    def __init__(self):
        required_parameters = [
            "name",
            "molar_mass",
            "vapor_pressure_strategy",
            "condensable",
            "concentration",
        ]
        BuilderABC.__init__(self, required_parameters)
        BuilderMolarMassMixin.__init__(self)
        BuilderConcentrationMixin.__init__(self, default_units="kg/m^3")
        self.name = None
        self.vapor_pressure_strategy = None
        self.condensable = None

    def set_name(self, name: Union[str, NDArray[np.str_]]):
        """Set the name of the gas species.

        Args:
            name : The name of the gas species.

        Returns:
            self : The GasSpeciesBuilder object.

        Examples:
            ``` py title="Set a single name"
            builder = GasSpeciesBuilder()
            builder.set_name("Oxygen")
            ```
        """
        self.name = name
        return self

    def set_vapor_pressure_strategy(
        self,
        strategy: Union[VaporPressureStrategy, list[VaporPressureStrategy]],
    ):
        """Set the vapor pressure strategy for the gas species.

        Args:
            strategy : The vapor pressure strategy for the gas species.

        Returns:
            self : The GasSpeciesBuilder object.

        Examples:
            ``` py title="Set a single vapor pressure strategy"
            builder = GasSpeciesBuilder()
            builder.set_vapor_pressure_strategy(
                ConstantVaporPressureStrategy(vapor_pressure=1.0)
            )
            ```
        """
        self.vapor_pressure_strategy = strategy
        return self

    def set_condensable(
        self,
        condensable: Union[bool, NDArray[np.bool_]],
    ):
        """Set the condensable bool of the gas species.

        Args:
            condensable : Whether the gas species is condensable.

        Returns:
            self : The GasSpeciesBuilder object.

        Examples:
            ``` py title="Set a single condensable bool"
            builder = GasSpeciesBuilder()
            builder.set_condensable(False)
            ```
        """
        self.condensable = condensable
        return self

    def build(self) -> GasSpecies:
        self.pre_build_check()
        return GasSpecies(
            name=self.name,  # type: ignore
            molar_mass=self.molar_mass,  # type: ignore
            vapor_pressure_strategy=(
                self.vapor_pressure_strategy
            ),  # type: ignore
            condensable=self.condensable,  # type: ignore
            concentration=self.concentration,  # type: ignore
        )


class PresetGasSpeciesBuilder(
    GasSpeciesBuilder,
):
    """Builder class for GasSpecies objects, allowing for a more fluent and
    readable creation of GasSpecies instances with optional parameters.
    """

    def __init__(self):
        GasSpeciesBuilder.__init__(self)
        self.name = "Preset100"
        self.molar_mass = 0.100  # kg/mol
        self.vapor_pressure_strategy = ConstantVaporPressureStrategy(
            vapor_pressure=1.0  # Pa
        )
        self.condensable = False
        self.concentration = 1.0

    def build(self) -> GasSpecies:
        self.pre_build_check()
        return GasSpecies(
            name=self.name,  # type: ignore
            molar_mass=self.molar_mass,  # type: ignore
            vapor_pressure_strategy=(
                self.vapor_pressure_strategy
            ),  # type: ignore
            condensable=self.condensable,  # type: ignore
            concentration=self.concentration,  # type: ignore
        )



"""Test of the ratio function in the coulomb_enhancement module."""

import numpy as np
from particula.particles.properties import coulomb_enhancement


def test_ratio_lower_limit():
    """Test the ratio function for ratio_lower_limit functionality."""
    # Test case where the calculated ratio is below the lower limit
    radius = 1e-9  # Very small radius to increase the Coulomb potential
    charge = 100   # Large charge to increase the Coulomb potential
    temperature = 298.15
    expected_result = -200  # Default ratio_lower_limit

    value = coulomb_enhancement.ratio(radius, charge, temperature)
    assert np.isclose(value, expected_result)

    # Test with a custom ratio_lower_limit
    custom_limit = -100
    expected_custom_result = custom_limit
    value_custom = coulomb_enhancement.ratio(
        radius, charge, temperature, ratio_lower_limit=custom_limit
    )
    assert np.isclose(value_custom, expected_custom_result)


def test_ratio():
    """Test the ratio function in the coulomb_enhancement module."""
    # Test case 1: Single radius and charge
    radius = 1.0
    charge = 2
    temperature = 298.15
    expected_result = -1.1203242349904997e-07
    value = coulomb_enhancement.ratio(radius, charge, temperature)
    assert np.isclose(value, expected_result)

    # Ensure existing test cases do not trigger the ratio_lower_limit
    # so they test the normal functionality of the ratio function.

    # Test case 2: Array of radii and charges
    radius = np.array([1.0, 2.0, 3.0])
    charge = np.array([-2, 3, 4])
    expected_result = np.array(
        [
            [-1.12032423e-07, 1.12032423e-07, 1.12032423e-07],
            [1.12032423e-07, -1.26036476e-07, -1.34438908e-07],
            [1.12032423e-07, -1.34438908e-07, -1.49376565e-07],
        ]
    )
    value = coulomb_enhancement.ratio(radius, charge, temperature)
    assert np.allclose(value, expected_result)


def test_kinetic():
    """Test the kinetic function in the coulomb_enhancement module."""
    # Test case 1: Positive coulomb potential
    coulomb_potential = 0.5
    expected_result = 1.5
    value = coulomb_enhancement.kinetic(coulomb_potential)
    assert np.isclose(value, expected_result)

    # Test case 2: Negative coulomb potential
    coulomb_potential = -0.5
    expected_result = 0.60653066
    value = coulomb_enhancement.kinetic(coulomb_potential)
    assert np.isclose(value, expected_result)


def test_continuum():
    """Test the continuum function in the coulomb_enhancement module."""
    # Test case 1: Non-zero coulomb potential
    coulomb_potential = 0.5
    expected_result = 1.27074704
    value = coulomb_enhancement.continuum(coulomb_potential)
    assert np.isclose(value, expected_result)

    # Test case 2: Zero coulomb potential
    coulomb_potential = 0.0
    expected_result = 1.0
    value = coulomb_enhancement.continuum(coulomb_potential)
    assert np.isclose(value, expected_result)



""" A centralized location for important, unchanged physics parameters.


    All constants are given in their base units. We use scipy constants.
"""

import scipy

BOLTZMANN_CONSTANT = scipy.constants.k

AVOGADRO_NUMBER = scipy.constants.Avogadro

# Gas constant in J mol^-1 K^-1 = m^2 kg mol^-1 s^-2 K^-1
# J = kg m^2 s^-2
GAS_CONSTANT = BOLTZMANN_CONSTANT * AVOGADRO_NUMBER

ELEMENTARY_CHARGE_VALUE = scipy.constants.elementary_charge

# Relative permittivity of air at approx.
# 296.15 K and 101325 Pa and 40% RH
# See https://www.osti.gov/servlets/purl/1504063
# Previously known as the "dielectric constant"
# Often denoted as epsilon
RELATIVE_PERMITTIVITY_AIR_ROOM = 1.000530569  # unitless
# At STP (273.15 K, 1 atm):
# see: https://en.wikipedia.org/wiki/Relative_permittivity
RELATIVE_PERMITTIVITY_AIR_STP = 1.00058986  # unitless

# select one of the two:
RELATIVE_PERMITTIVITY_AIR = RELATIVE_PERMITTIVITY_AIR_ROOM

# Permittivity of free space in F/m
# Also known as the electric constant, permittivity of free space
# Often denoted by epsilon_0
VACUUM_PERMITTIVITY = scipy.constants.epsilon_0

ELECTRIC_PERMITTIVITY = RELATIVE_PERMITTIVITY_AIR * VACUUM_PERMITTIVITY

# These values are used to calculate the dynamic viscosity of air
# Here, REF temperature and viscosity are at STP:
# Standard temperature and pressure (273.15 K and 101325 Pa)
REF_VISCOSITY_AIR_STP = 1.716e-5  # Pa*s
REF_TEMPERATURE_STP = 273.15  # K
SUTHERLAND_CONSTANT = 110.4  # K
MOLECULAR_WEIGHT_AIR = 0.0289644  # kg/mol

STANDARD_GRAVITY = scipy.constants.g



"""
Convert the oxygen to carbon ratio and molar mass ratio
to OH equivalent.

Gorkowski, K., Preston, T. C., &#38; Zuend, A. (2019).
Relative-humidity-dependent organic aerosol thermodynamics
Via an efficient reduced-complexity model.
Atmospheric Chemistry and Physics
https://doi.org/10.5194/acp-19-13383-2019
"""

from typing import Optional, Union, Tuple
import numpy as np
from numpy.typing import NDArray


def convert_to_oh_equivalent(
    oxygen2carbon: Union[float, NDArray[np.float64]],
    molar_mass_ratio: Union[float, NDArray[np.float64]],
    functional_group: Optional[Union[list[str], str]] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Just a pass through now, but will
    add the oh equivalent conversion.

    Args:
        - oxygen2carbon : The oxygen to carbon ratio.
        - molar_mass_ratio : The molar mass ratio of water to organic
          matter.
        - functional_group : Optional functional group(s) of the organic
          compound, if applicable.

    Returns:
        - A tuple containing the converted oxygen to carbon ratio and
          molar mass ratio.
    """

    if functional_group is None:
        return oxygen2carbon, molar_mass_ratio
    if functional_group == "alcohol":
        return oxygen2carbon + 1, molar_mass_ratio + 16  # fix this from SI
    raise ValueError("BAT functional group not recognized")



""" calculating reduced quantity.

    reduced_quantity =
        quantity_1 * quantity_2 / (quantity_1 + quantity_2)
"""

import logging
from typing import Union
from numpy.typing import NDArray
import numpy as np

logger = logging.getLogger("particula")  # get instance of logger


def reduced_value(
    alpha: Union[float, NDArray[np.float64]],
    beta: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """
    Returns the reduced value of two parameters, calculated as:
    reduced_value = alpha * beta / (alpha + beta)

    This formula calculates an "effective inertial" quantity,
    allowing two-body problems to be solved as if they were one-body problems.

    Args:
        - alpha : The first parameter (scalar or array).
        - beta : The second parameter (scalar or array).

    Returns:
        - A value or array of the same dimension as the input parameters.
            Returns zero where alpha + beta equals zero to handle division
            by zero gracefully.

    Raises:
        - ValueError : If alpha and beta are arrays and their shapes do
            not match.
    """
    # Ensure input compatibility, especially when both are arrays
    if (
        isinstance(alpha, np.ndarray)
        and isinstance(beta, np.ndarray)
        and (alpha.shape != beta.shape)
    ):
        logger.error("The shapes of alpha and beta must be identical.")
        raise ValueError("The shapes of alpha and beta must be identical.")

    # Calculation of the reduced value, with safety against division by zero
    denominator = alpha + beta
    # Using np.errstate to suppress divide by zero warnings
    with np.errstate(divide="ignore", invalid="ignore"):
        result = np.where(denominator != 0, alpha * beta / denominator, 0)
    return result


def reduced_self_broadcast(
    alpha_array: NDArray[np.float64],
) -> NDArray[np.float64]:
    """
    Returns the reduced value of an array with itself, broadcasting the
    array into a matrix and calculating the reduced value of each element pair.
    reduced_value = alpha_matrix * alpha_matrix_Transpose
                    / (alpha_matrix + alpha_matrix_Transpose)

    Args:
        - alpha_array : The array to be broadcast and reduced.

    Returns:
        - A square matrix of the reduced values.
    """
    # Use broadcasting to create matrix and its transpose
    alpha_matrix = alpha_array[:, np.newaxis]
    alpha_matrix_transpose = alpha_array[np.newaxis, :]
    denominator = alpha_matrix + alpha_matrix_transpose
    # Perform element-wise multiplication and division with np.errstate
    with np.errstate(divide="ignore", invalid="ignore"):
        result = np.where(
            denominator != 0,
            alpha_matrix * alpha_matrix_transpose / denominator,
            0,
        )
    return result



"""
This module contains functions to calculates the activity coefficients
at a fixed water activity or the water activity cross point.

Gorkowski, K., Preston, T. C., Zuend, A. (2019).
Relative-humidity-dependent organic aerosol thermodynamics
Via an efficient reduced-complexity model.
Atmospheric Chemistry and Physics
https://doi.org/10.5194/acp-19-13383-2019
"""

from typing import Union, Optional, Tuple
import numpy as np
from numpy.typing import NDArray
from particula.activity.bat_coefficients import (
    LOWEST_ORGANIC_MOLE_FRACTION,
    INTERPOLATE_WATER_FIT,
)
from particula.activity.activity_coefficients import bat_activity_coefficients
from particula.activity.species_density import organic_density_estimate
from particula.activity import phase_separation


def biphasic_water_activity_point(
    oxygen2carbon: Union[float, NDArray[np.float64]],
    hydrogen2carbon: Union[float, NDArray[np.float64]],
    molar_mass_ratio: Union[float, NDArray[np.float64]],
    functional_group: Optional[Union[list[str], str]] = None,
) -> np.ndarray:
    """
    This function computes the biphasic to single phase
    water activity (RH*100).

    Args:
        - oxygen2carbon : The oxygen to carbon ratio.
        - hydrogen2carbon : The hydrogen to carbon ratio.
        - molar_mass_ratio : The molar mass ratio of water to organic
          matter.
        - functional_group : Optional functional group(s) of the organic
          compound, if applicable.

    Returns:
        - The RH cross point array.
    """
    oxygen2carbon = np.asarray(oxygen2carbon, dtype=np.float64)
    hydrogen2carbon = np.asarray(hydrogen2carbon, dtype=np.float64)
    molar_mass_ratio = np.asarray(molar_mass_ratio, dtype=np.float64)
    if oxygen2carbon.ndim == 0:
        oxygen2carbon = np.expand_dims(oxygen2carbon, axis=0)
    if hydrogen2carbon.ndim == 0:
        hydrogen2carbon = np.expand_dims(hydrogen2carbon, axis=0)
    if molar_mass_ratio.ndim == 0:
        molar_mass_ratio = np.expand_dims(molar_mass_ratio, axis=0)

    water_activity_cross_point = np.zeros_like(oxygen2carbon)

    interpolate_step_numb = 200
    mole_frac = np.logspace(-6, 0, interpolate_step_numb + 1)

    for i, _ in enumerate(oxygen2carbon):
        density = organic_density_estimate(
            molar_mass_ratio[i],
            oxygen2carbon[i],
            hydrogen2carbon[i],
            mass_ratio_convert=True,
        )
        activities = bat_activity_coefficients(
            molar_mass_ratio=molar_mass_ratio[i],
            organic_mole_fraction=mole_frac,
            oxygen2carbon=oxygen2carbon[i],
            density=density,
            functional_group=functional_group,
        )

        if np.isnan(activities[0]).any():
            raise ValueError("water activity is NaN, check inputs")

        phase_check = phase_separation.find_phase_separation(
            activities[0], activities[1]
        )

        if phase_check["phase_sep_check"] == 1:
            water_activity_cross_point[i] = phase_check["upper_seperation"]
        else:
            water_activity_cross_point[i] = 0

    water_activity_cross_point[water_activity_cross_point < 0] = 0
    water_activity_cross_point[water_activity_cross_point > 1] = 1

    return water_activity_cross_point


def fixed_water_activity(
    water_activity: Union[float, NDArray[np.float64]],
    molar_mass_ratio: Union[float, NDArray[np.float64]],
    oxygen2carbon: Union[float, NDArray[np.float64]],
    density: Union[float, NDArray[np.float64]],
) -> Tuple[
    Union[float, NDArray[np.float64]],
    Union[float, NDArray[np.float64]],
    Union[float, NDArray[np.float64]],
]:
    """
    Calculate the activity coefficients of water and organic matter in
    organic-water mixtures.

    This function assumes a fixed water activity value (e.g., RH = 75%
    corresponds to 0.75 water activity in equilibrium).
    It calculates the activity coefficients for different phases and
    determines phase separations if they occur.

    Args:
        - water_activity : An array of water activity values.
        - molar_mass_ratio : Array of molar mass ratios of the components.
        - oxygen2carbon : Array of oxygen-to-carbon ratios.
        - density : Array of densities of the mixture, in kg/m^3.

    Returns:
        - A tuple containing the activity coefficients for alpha and beta
          phases, and the q_alpha (phase separation) value.
          If no phase separation occurs, the beta phase values are None.
    """
    # pylint: disable=too-many-locals

    # check types
    water_activity = np.asarray(water_activity, dtype=np.float64)
    molar_mass_ratio = np.asarray(molar_mass_ratio, dtype=np.float64)
    oxygen2carbon = np.asarray(oxygen2carbon, dtype=np.float64)
    density = np.asarray(density, dtype=np.float64)

    # must have activity of water in increasing order
    if water_activity.size > 1 and water_activity[0] > water_activity[-1]:
        water_activity = np.flip(water_activity)
        flip = True
    else:
        flip = False

    organic_mole_fraction_array = np.linspace(
        1,
        LOWEST_ORGANIC_MOLE_FRACTION,
        INTERPOLATE_WATER_FIT,
        dtype=np.float64,
    )

    # activity calculation
    activities = bat_activity_coefficients(
        molar_mass_ratio=molar_mass_ratio,
        organic_mole_fraction=organic_mole_fraction_array,
        oxygen2carbon=oxygen2carbon,
        density=density,
    )
    # find phase separation
    phase_check = phase_separation.find_phase_separation(
        activities[0], activities[1]
    )
    # ensure water activity type is float
    activities_water = np.asarray(activities[0], dtype=np.float64)
    if phase_check["phase_sep_check"] == 0:
        alpha_organic_mole_fraction = np.interp(
            xp=activities_water,
            fp=organic_mole_fraction_array,
            x=water_activity,
            left=1.0,
            right=LOWEST_ORGANIC_MOLE_FRACTION,
        )
        # activity calculation for alpha phase
        activities_alpha = bat_activity_coefficients(
            molar_mass_ratio=molar_mass_ratio,
            organic_mole_fraction=alpha_organic_mole_fraction,
            oxygen2carbon=oxygen2carbon,
            density=density,
        )
        activities_beta = None
        q_alpha = np.ones(water_activity.size)
        # change back to original order
        if flip:
            activities_alpha = np.flip(activities_alpha)
            q_alpha = np.flip(q_alpha)
        return (activities_alpha, activities_beta, q_alpha)

    # else phase separation occurs
    # split the activities into alpha and beta phases

    # alpha water rich phase
    alpha_water_activity = activities_water[
        phase_check["upper_seperation_index"]:
    ]
    alpha_organic_mole_fraction = organic_mole_fraction_array[
        phase_check["upper_seperation_index"]:
    ]
    # beta organic rich phase
    beta_water_activity = activities_water[
        : phase_check["matching_upper_seperation_index"]
    ]
    beta_organic_mole_fraction = organic_mole_fraction_array[
        : phase_check["matching_upper_seperation_index"]
    ]

    # find the water activity of the alpha phase
    alpha_organic_mole_fraction_interp = np.interp(
        xp=alpha_water_activity,
        fp=alpha_organic_mole_fraction,
        x=water_activity,
        left=np.nan,
        right=LOWEST_ORGANIC_MOLE_FRACTION,
    )
    # find the water activity of the beta phase
    beta_organic_mole_fraction_interp = np.interp(
        xp=beta_water_activity,
        fp=beta_organic_mole_fraction,
        x=water_activity,
        left=1,
        right=np.nan,
    )
    # calculate the activity coefficients for the alpha phase
    activities_alpha = bat_activity_coefficients(
        molar_mass_ratio=molar_mass_ratio,
        organic_mole_fraction=alpha_organic_mole_fraction_interp,
        oxygen2carbon=oxygen2carbon,
        density=density,
    )
    # calculate the activity coefficients for the beta phase
    activities_beta = bat_activity_coefficients(
        molar_mass_ratio=molar_mass_ratio,
        organic_mole_fraction=beta_organic_mole_fraction_interp,
        oxygen2carbon=oxygen2carbon,
        density=density,
    )
    q_alpha = phase_separation.q_alpha(
        seperation_activity=phase_check["upper_seperation"],
        activities=water_activity,
    )
    # change back to original order
    if flip:
        activities_alpha = np.flip(activities_alpha)
        activities_beta = np.flip(activities_beta)
        q_alpha = np.flip(q_alpha)

    return (activities_alpha, activities_beta, q_alpha)



"""
Vapor pressure modules for calculating the vapor pressure of a gas.
"""

from typing import Union
import numpy as np
from numpy.typing import NDArray
from particula.util.constants import GAS_CONSTANT


# Antoine equation function
def antoine_vapor_pressure(
    a: Union[float, NDArray[np.float64]],
    b: Union[float, NDArray[np.float64]],
    c: Union[float, NDArray[np.float64]],
    temperature: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate vapor pressure using the Antoine equation.

    Args:
        a, b, c: Antoine equation parameters.
        temperature: Temperature in Kelvin.

    Returns:
        Vapor pressure in Pascals.

    References:
        - Equation: log10(P) = a - b / (T - c)
        - https://en.wikipedia.org/wiki/Antoine_equation (but in Kelvin)
        - Kelvin form:
            https://onlinelibrary.wiley.com/doi/pdf/10.1002/9781118135341.app1
    """
    vapor_pressure_log = a - (b / (temperature - c))
    vapor_pressure = 10**vapor_pressure_log
    return vapor_pressure * 133.32238741499998  # Convert mmHg to Pa


# Clausius-Clapeyron equation function
def clausius_clapeyron_vapor_pressure(
    latent_heat: Union[float, NDArray[np.float64]],
    temperature_initial: Union[float, NDArray[np.float64]],
    pressure_initial: Union[float, NDArray[np.float64]],
    temperature: Union[float, NDArray[np.float64]],
    gas_constant: float = GAS_CONSTANT,
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate vapor pressure using Clausius-Clapeyron equation.

    Args:
        latent_heat: Latent heat of vaporization in J/mol.
        temperature_initial: Initial temperature in Kelvin.
        pressure_initial: Initial vapor pressure in Pascals.
        temperature: Final temperature in Kelvin.
        gas_constant: gas constant (default is 8.314 J/(mol·K)).

    Returns:
        Pure vapor pressure in Pascals.

    References:
        - https://en.wikipedia.org/wiki/Clausius%E2%80%93Clapeyron_relation
    """
    return pressure_initial * np.exp(
        (latent_heat / gas_constant)
        * (1 / temperature_initial - 1 / temperature)
    )


# Buck equation function
def buck_vapor_pressure(
    temperature: Union[float, NDArray[np.float64]]
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate vapor pressure using the Buck equation for water vapor.

    Args:
        temperature: Temperature in Kelvin.

    Returns:
        Vapor pressure in Pascals.

    References:
        - Buck, A. L., 1981: New Equations for Computing Vapor Pressure and
            Enhancement Factor. J. Appl. Meteor. Climatol., 20, 1527-1532,
            https://doi.org/10.1175/1520-0450(1981)020<1527:NEFCVP>2.0.CO;2.
        - https://en.wikipedia.org/wiki/Arden_Buck_equation
    """
    temp = np.array(temperature) - 273.15  # Convert to Celsius
    temp_below_freezing = temp < 0.0
    temp_above_freezing = temp >= 0.0

    vapor_pressure_below_freezing = (
        6.1115 * np.exp((23.036 - temp / 333.7) * temp / (279.82 + temp)) * 100
    )  # hPa to Pa
    vapor_pressure_above_freezing = (
        6.1121 * np.exp((18.678 - temp / 234.5) * temp / (257.14 + temp)) * 100
    )  # hPa to Pa

    return (
        vapor_pressure_below_freezing * temp_below_freezing
        + vapor_pressure_above_freezing * temp_above_freezing
    )



"""
Particle inertia time calculation.
"""
from typing import Union
from numpy.typing import NDArray
import numpy as np

from particula.util.validate_inputs import validate_inputs


@validate_inputs(
    {
        "particle_radius": "positive",
        "particle_density": "positive",
        "fluid_density": "positive",
        "kinematic_viscosity": "positive",
        "relative_velocity": "positive",
    }
)
def get_particle_inertia_time(
    particle_radius: Union[float, NDArray[np.float64]],
    particle_density: Union[float, NDArray[np.float64]],
    fluid_density: Union[float, NDArray[np.float64]],
    kinematic_viscosity: Union[float, NDArray[np.float64]],
    relative_velocity: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the particle inertia time.

    The particle inertia time (τ_p) represents the response time of a particle
    to velocity changes in the surrounding fluid and is given by:

        τ_p = (2 / 9) * (ρ_p / ρ_f) * (r² / (ν f(Re_p)))

    - τ_p : Particle inertia time [s]
    - r (particle_radius) : Particle radius [m]
    - ρ_p (particle_density) : Particle (e.g., droplet, dust) density [kg/m³]
    - ρ_f (fluid_density) : Density of the surrounding fluid (e.g., air, water)
        [kg/m³]
    - ν (kinematic_viscosity) : Kinematic viscosity of the fluid [m²/s]
    - f(Re_p) : Drag correction factor based on the particle Reynolds number
        [-]

    The drag correction factor f(Re_p) is given by:

        f(Re_p) = 1 + 0.15 Re_p^(0.687)

    where the particle Reynolds number is:

        Re_p = (2 r v) / v

    - v (relative_velocity) : Relative velocity between particle and fluid
        [m/s]

    Arguments:
    ----------
        - particle_radius : Radius of the particle [m]
        - particle_density : Density of the particle [kg/m³]
        - fluid_density : Density of the surrounding fluid [kg/m³]
        - kinematic_viscosity : Kinematic viscosity of the fluid [m²/s]
        - relative_velocity : Relative velocity between particle and fluid
            [m/s]

    Returns:
    --------
        - Particle inertia time [s]

    References:
    ----------
        - Ayala, O., Rosa, B., & Wang, L. P. (2008). Effects of turbulence on
        the geometric collision rate of sedimenting droplets. Part 2. Theory
        and parameterization. New Journal of Physics, 10.
        https://doi.org/10.1088/1367-2630/10/7/075016
    """
    re_p = (2 * particle_radius * relative_velocity) / kinematic_viscosity
    drag_correction = 1 + 0.15 * re_p**0.687
    return (
        (2 / 9)
        * (particle_density / fluid_density)
        * (particle_radius**2 / (kinematic_viscosity * drag_correction))
    )



"""Machine max or min overflow protection."""

from numpy.typing import ArrayLike
import numpy as np


MIN_POSITIVE_VALUE = np.nextafter(0, 1, dtype=np.float64)
MAX_POSITIVE_VALUE = np.finfo(np.float64).max
MAX_NEGATIVE_VALUE = np.finfo(np.float64).min


def safe_exp(value: ArrayLike) -> np.ndarray:
    """
    Compute the exponential of each element in the input array, with limits
    to prevent overflow based on machine precision.

    Args:
        value (ArrayLike): Input array.

    Returns:
        np.ndarray: Exponential of the input array with overflow protection.
    """
    value = np.asarray(value, dtype=np.float64)
    max_exp_input = np.log(np.finfo(value.dtype).max)
    return np.exp(np.clip(value, None, max_exp_input))


def safe_log(value: ArrayLike) -> np.ndarray:
    """
    Compute the natural logarithm of each element in the input array, with
    limits to prevent underflow based on machine precision.

    Args:
        value (ArrayLike): Input array.

    Returns:
        np.ndarray: Natural logarithm of the input array with underflow
        protection.
    """
    value = np.asarray(value, dtype=np.float64)
    min_positive_value = np.nextafter(0, 1, dtype=value.dtype)
    return np.log(np.clip(value, min_positive_value, None))


def safe_log10(value: ArrayLike) -> np.ndarray:
    """
    Compute the base 10 logarithm of each element in the input array, with
    limits to prevent underflow based on machine precision.

    Args:
        value (ArrayLike): Input array.

    Returns:
        np.ndarray: Base 10 logarithm of the input array with underflow
        protection.
    """
    value = np.asarray(value, dtype=np.float64)
    min_positive_value = np.nextafter(0, 1, dtype=value.dtype)
    return np.log10(np.clip(value, min_positive_value, None))



"""Functions for calculating the partial pressure of a gas."""

from typing import Union
from numpy.typing import NDArray
import numpy as np

from particula.util.constants import GAS_CONSTANT


def calculate_partial_pressure(
    concentration: Union[float, NDArray[np.float64]],
    molar_mass: Union[float, NDArray[np.float64]],
    temperature: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the partial pressure of a gas from its concentration, molar mass,
    and temperature.

    Parameters:
    - concentration (float): Concentration of the gas in kg/m^3.
    - molar_mass (float): Molar mass of the gas in kg/mol.
    - temperature (float): Temperature in Kelvin.

    Returns:
    - float: Partial pressure of the gas in Pascals (Pa).
    """
    # Input validation
    if np.any(concentration < 0):
        raise ValueError("Concentration must be positive")
    if np.any(molar_mass <= 0):
        raise ValueError("Molar mass must be positive")
    if np.any(temperature <= 0):
        raise ValueError("Temperature must be positive")

    # Calculate the partial pressure
    return (concentration * float(GAS_CONSTANT) * temperature) / molar_mass


def calculate_saturation_ratio(
    partial_pressure: Union[float, NDArray[np.float64]],
    pure_vapor_pressure: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the saturation ratio of the gas at a given pressure and
    temperature.

    Args:
        partial_pressure: Pressure in Pascals.
        pure_vapor_pressure: Pure vapor pressure of the gas in Pascals.

    Returns:
        saturation_ratio: The saturation ratio of the gas.
    """
    return partial_pressure / pure_vapor_pressure



"""
Test the turbulent shear kernel functions.
"""

import numpy as np
import pytest
from particula.dynamics.coagulation.turbulent_shear_kernel import (
    saffman_turner_1956,
    saffman_turner_1956_via_system_state,
)
from particula.gas.properties.kinematic_viscosity import (
    get_kinematic_viscosity_via_system_state,
)


def test_turbulent_shear_kernel_single_value():
    """
    Test turbulent_shear_kernel with single float inputs for particle
    diameters, turbulent kinetic energy, and kinematic viscosity.
    """
    diameter_particle = np.array([1e-6, 2e-6])  # example diameters [m]
    turbulent_kinetic_energy = (
        1.0e-4  # example turbulent kinetic energy [m^2/s^2]
    )
    kinematic_viscosity = 1.5e-5  # example kinematic viscosity [m^2/s]
    expected_kernel = (
        np.pi * turbulent_kinetic_energy / (120 * kinematic_viscosity)
    ) ** 0.5 * (
        diameter_particle[:, np.newaxis] + diameter_particle[np.newaxis, :]
    ) ** 3
    value = saffman_turner_1956(
        diameter_particle/2, turbulent_kinetic_energy, kinematic_viscosity
    )
    np.testing.assert_allclose(value, expected_kernel, rtol=1e-6)


def test_turbulent_shear_kernel_via_system_state():
    """
    Test turbulent_shear_kernel_via_system_state with system state inputs.
    """
    particle_radius = np.array([1e-6, 2e-6])  # example diameters [m]
    turbulent_kinetic_energy = (
        1.0e-4  # example turbulent kinetic energy [m^2/s^2]
    )
    temperature = 298  # example temperature [K]
    fluid_density = 1.2  # example fluid density [kg/m^3]
    kinematic_viscosity = get_kinematic_viscosity_via_system_state(
        temperature=temperature, fluid_density=fluid_density
    )
    expected_kernel = saffman_turner_1956(
        particle_radius, turbulent_kinetic_energy, kinematic_viscosity
    )
    value = saffman_turner_1956_via_system_state(
        particle_radius, turbulent_kinetic_energy, temperature, fluid_density
    )
    np.testing.assert_allclose(value, expected_kernel, rtol=1e-6)


def test_turbulent_shear_kernel_input_validation():
    """
    Ensure that providing incorrect input types to turbulent_shear_kernel
    raises a TypeError.
    """
    with pytest.raises(TypeError):
        saffman_turner_1956("not a number", "not a number", "not a number")


def test_turbulent_shear_kernel_via_system_state_input_validation():
    """
    Ensure that providing incorrect input types to
    turbulent_shear_kernel_via_system_state raises a TypeError.
    """
    with pytest.raises(TypeError):
        saffman_turner_1956_via_system_state(
            "not a number", "not a number", "not a number", "not a number"
        )



""" Module contains the function for calculating the mean thermal speed
of particles in a fluid."""

from typing import Union
from numpy.typing import NDArray
import numpy as np

from particula.util.constants import BOLTZMANN_CONSTANT


def mean_thermal_speed(
    mass: Union[float, NDArray[np.float64]],
    temperature: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Returns the particles mean thermal speed. Due to the the impact
    of air molecules on the particles, the particles will have a mean
    thermal speed.

    Args
    ----
    mass : The per particle mass of the particles [kg].
    temperature : The temperature of the air [K].

    Returns
    -------
    The mean thermal speed of the particles [m/s].

    References
    ----------
    Seinfeld, J. H., & Pandis, S. N. (2016). Atmospheric chemistry and
    physics, Section 9.5.3 Mean Free Path of an Aerosol Particle Equation 9.87.
    """
    return np.sqrt((8 * BOLTZMANN_CONSTANT * temperature) / (np.pi * mass))



"""Test the surface module."""

import numpy as np
import pytest
from particula.particles.surface_strategies import (
    SurfaceStrategyMass,
    SurfaceStrategyMolar,
    SurfaceStrategyVolume,
)


# Test MolarSurfaceStrategy
def test_molar_surface_strategy():
    """Test MolarSurfaceStrategy class."""
    surface_tension = np.array([0.072, 0.05])  # water, oil
    density = np.array([1000, 800])  # water, oil
    molar_mass = np.array([0.01815, 0.03])  # water, oil

    strategy = SurfaceStrategyMolar(surface_tension, density, molar_mass)

    # Test effective surface tension
    mass_concentration = np.array([100, 200])  # water, oil
    expected_surface_tension = 0.0599547511312
    assert strategy.effective_surface_tension(
        mass_concentration
    ) == pytest.approx(expected_surface_tension)

    # Test effective density
    expected_density = 890.497737556561
    assert strategy.effective_density(mass_concentration) == pytest.approx(
        expected_density
    )

    # Test kelvin_radius
    molar_mass_water = molar_mass[0]
    expected_kelvin_radius = (
        2 * expected_surface_tension * molar_mass_water
    ) / (8.314 * 298 * expected_density)
    assert strategy.kelvin_radius(
        molar_mass_water, mass_concentration, 298
    ) == pytest.approx(expected_kelvin_radius, rel=1e-3)

    # Test kelvin_term
    radius = 1e-6
    expected_kelvin_term = np.exp(expected_kelvin_radius / radius)
    assert strategy.kelvin_term(
        radius, molar_mass_water, mass_concentration, 298
    ) == pytest.approx(expected_kelvin_term)


# Test MassSurfaceStrategy
def test_mass_surface_strategy():
    """Test MassSurfaceStrategy class."""
    surface_tension = np.array([0.072, 0.05])  # water, oil
    density = np.array([1000, 800])  # water, oil

    strategy = SurfaceStrategyMass(surface_tension, density)

    # Test effective surface tension
    mass_concentration = np.array([100, 200])  # water, oil
    expected_surface_tension = surface_tension[0] * 100 / (
        100 + 200
    ) + surface_tension[1] * 200 / (100 + 200)
    assert strategy.effective_surface_tension(
        mass_concentration
    ) == pytest.approx(expected_surface_tension)

    # Test effective density
    expected_density = density[0] * 100 / (100 + 200) + density[1] * 200 / (
        100 + 200
    )
    assert strategy.effective_density(mass_concentration) == pytest.approx(
        expected_density
    )

    # Test kelvin_radius
    expected_kelvin_radius = 9.691952451532837e-10
    assert strategy.kelvin_radius(
        0.01815, mass_concentration, 298
    ) == pytest.approx(expected_kelvin_radius, rel=1e-6)

    # Test kelvin_term
    radius = 1e-6
    expected_kelvin_term = np.exp(expected_kelvin_radius / radius)
    assert strategy.kelvin_term(
        radius, 0.01815, mass_concentration, 298
    ) == pytest.approx(expected_kelvin_term, rel=1e-6)


# Test VolumeSurfaceStrategy
def test_volume_surface_strategy():
    """Test VolumeSurfaceStrategy class."""
    surface_tension = np.array([0.072, 0.05])  # water, oil
    density = np.array([1000, 800])  # water, oil

    strategy = SurfaceStrategyVolume(surface_tension, density)

    # Test effective surface tension
    mass_concentration = np.array([100, 200])  # water, oil
    expected_surface_tension = 0.05628571428571429
    assert strategy.effective_surface_tension(
        mass_concentration
    ) == pytest.approx(expected_surface_tension)

    # Test effective density
    expected_density = 857.1428571428572
    assert strategy.effective_density(mass_concentration) == pytest.approx(
        expected_density
    )

    # Test kelvin_radius
    expected_kelvin_radius = 9.62057760789752e-10
    assert strategy.kelvin_radius(
        0.01815, mass_concentration, 298
    ) == pytest.approx(expected_kelvin_radius)

    # Test kelvin_term
    radius = 1e-6
    expected_kelvin_term = np.exp(expected_kelvin_radius / radius)
    assert strategy.kelvin_term(
        radius, 0.01815, mass_concentration, 298
    ) == pytest.approx(expected_kelvin_term)



"""Testing the distribution factories, for the distribution strategies.

The factories are tested to ensure that the correct distribution strategy is
created based on the factory used.

The Strategy is tested independently.
"""

import pytest
from particula.particles.distribution_factories import (
    DistributionFactory,
)
from particula.particles.distribution_strategies import (
    MassBasedMovingBin,
    RadiiBasedMovingBin,
    SpeciatedMassMovingBin,
    ParticleResolvedSpeciatedMass,
)


def test_mass_based_moving_bin():
    """Test factory function for mass-based moving bin strategy."""
    strategy = DistributionFactory().get_strategy("mass_based_moving_bin")
    assert isinstance(strategy, MassBasedMovingBin)


def test_radii_based_moving_bin():
    """Test factory function for radii-based moving bin strategy."""
    strategy = DistributionFactory().get_strategy("radii_based_moving_bin")
    assert isinstance(strategy, RadiiBasedMovingBin)


def test_speciated_mass_moving_bin():
    """Test factory function for speciated mass moving bin strategy."""
    strategy = DistributionFactory().get_strategy("speciated_mass_moving_bin")
    assert isinstance(strategy, SpeciatedMassMovingBin)


def test_particle_resolved_speciated_mass():
    """Test factory function for particle resolved speciated mass strategy."""
    strategy = DistributionFactory().get_strategy(
        "particle_resolved_speciated_mass"
    )
    assert isinstance(strategy, ParticleResolvedSpeciatedMass)


def test_invalid_strategy():
    """Test factory function for invalid type."""
    with pytest.raises(ValueError) as excinfo:
        DistributionFactory().get_strategy("invalid_type")
    assert "Unknown strategy type: invalid_type" in str(excinfo.value)



"""
Tests for the velocity_correlation_terms module.
"""

import pytest
import numpy as np
from particula.dynamics.coagulation.turbulent_dns_kernel.velocity_correlation_terms_ao2008 import (
    compute_z,
    compute_beta,
    compute_b1,
    compute_b2,
    compute_c1,
    compute_c2,
    compute_d1,
    compute_d2,
    compute_e1,
    compute_e2,
)


def test_compute_z():
    """
    Test compute_z with scalar and array inputs.
    """
    lagrangian_integral_time = np.array([1.0, 2.0, 3.0])  # [s]
    eulerian_integral_length = np.array([10.0, 20.0, 30.0])  # [m]

    expected = lagrangian_integral_time / eulerian_integral_length
    result = compute_z(lagrangian_integral_time, eulerian_integral_length)

    assert np.allclose(result, expected), f"Expected {expected}, got {result}"


def test_compute_beta():
    """
    Test compute_beta with scalar and array inputs.
    """
    taylor_microscale = np.array([1.0, 2.0, 3.0])  # [m]
    eulerian_integral_length = np.array([10.0, 20.0, 30.0])  # [m]

    expected = (np.sqrt(2) * taylor_microscale) / eulerian_integral_length
    result = compute_beta(taylor_microscale, eulerian_integral_length)

    assert np.allclose(result, expected), f"Expected {expected}, got {result}"


def test_compute_b1():
    """
    Test compute_b1 with valid inputs.
    """
    z = np.array([0.1, 0.2, 0.3])  # [-]
    expected = (1 + np.sqrt(1 - 2 * z**2)) / (2 * np.sqrt(1 - 2 * z**2))
    result = compute_b1(z)

    assert np.allclose(result, expected), f"Expected {expected}, got {result}"


def test_compute_b2():
    """
    Test compute_b2 with valid inputs.
    """
    z = np.array([0.1, 0.2, 0.3])  # [-]
    expected = (1 - np.sqrt(1 - 2 * z**2)) / (2 * np.sqrt(1 - 2 * z**2))
    result = compute_b2(z)

    assert np.allclose(result, expected), f"Expected {expected}, got {result}"


def test_compute_c1():
    """
    Test compute_c1 with valid inputs.
    """
    z = np.array([0.1, 0.2, 0.3])  # [-]
    lagrangian_integral_time = np.array([10.0, 20.0, 30.0])  # [s]

    expected = ((1 + np.sqrt(1 - 2 * z**2)) * lagrangian_integral_time) / 2
    result = compute_c1(z, lagrangian_integral_time)

    assert np.allclose(result, expected), f"Expected {expected}, got {result}"


def test_compute_c2():
    """
    Test compute_c2 with valid inputs.
    """
    z = np.array([0.1, 0.2, 0.3])  # [-]
    lagrangian_integral_time = np.array([10.0, 20.0, 30.0])  # [s]

    expected = ((1 - np.sqrt(1 - 2 * z**2)) * lagrangian_integral_time) / 2
    result = compute_c2(z, lagrangian_integral_time)

    assert np.allclose(result, expected), f"Expected {expected}, got {result}"


def test_compute_d1():
    """
    Test compute_d1 with valid inputs.
    """
    beta = np.array([0.1, 0.2, 0.3])  # [-]

    expected = (1 + np.sqrt(1 - 2 * beta**2)) / (2 * np.sqrt(1 - 2 * beta**2))
    result = compute_d1(beta)

    assert np.allclose(result, expected), f"Expected {expected}, got {result}"


def test_compute_d2():
    """
    Test compute_d2 with valid inputs.
    """
    beta = np.array([0.1, 0.2, 0.3])  # [-]

    expected = (1 - np.sqrt(1 - 2 * beta**2)) / (2 * np.sqrt(1 - 2 * beta**2))
    result = compute_d2(beta)

    assert np.allclose(result, expected), f"Expected {expected}, got {result}"


def test_compute_e1():
    """
    Test compute_e1 with valid inputs.
    """
    beta = np.array([0.1, 0.2, 0.3])  # [-]
    eulerian_integral_length = np.array([10.0, 20.0, 30.0])  # [m]

    expected = ((1 + np.sqrt(1 - 2 * beta**2)) * eulerian_integral_length) / 2
    result = compute_e1(beta, eulerian_integral_length)

    assert np.allclose(result, expected), f"Expected {expected}, got {result}"


def test_compute_e2():
    """
    Test compute_e2 with valid inputs.
    """
    beta = np.array([0.1, 0.2, 0.3])  # [-]
    eulerian_integral_length = np.array([10.0, 20.0, 30.0])  # [m]

    expected = ((1 - np.sqrt(1 - 2 * beta**2)) * eulerian_integral_length) / 2
    result = compute_e2(beta, eulerian_integral_length)

    assert np.allclose(result, expected), f"Expected {expected}, got {result}"


def test_invalid_inputs():
    """
    Ensure validation errors are raised for invalid inputs.
    """
    with pytest.raises(ValueError):
        compute_z(-1, 10)  # Negative lagrangian_integral_time

    with pytest.raises(ValueError):
        compute_beta(1, -10)  # Negative length scale

    with pytest.raises(ValueError):
        compute_b1(-0.5)  # Invalid z value

    with pytest.raises(ValueError):
        compute_d1(-0.5)  # Invalid beta value

    with pytest.raises(ValueError):
        compute_e1(0.5, -10)  # Negative length scale


def test_edge_cases():
    """
    Test edge cases such as very small and very large values.
    """
    lagrangian_integral_time = np.array(
        [1e-6, 1e-3, 10.0]
    )  # Very small and large inertia values
    eulerian_integral_length = np.array(
        [1e-6, 1e-3, 10.0]
    )  # Very small and large values

    z_values = compute_z(lagrangian_integral_time, eulerian_integral_length)
    beta_values = compute_beta(lagrangian_integral_time, eulerian_integral_length)

    assert np.all(np.isfinite(z_values)), "Expected all z values to be finite"
    assert np.all(
        np.isfinite(beta_values)
    ), "Expected all beta values to be finite"



"""
Tests for Converting Units
"""
import unittest
from particula.util.converting.units import convert_units
# flake8: noqa
try:
    import pint  # pylint: disable=unused-import
except ImportError:
    IS_PINT_AVAILABE = False
else:
    IS_PINT_AVAILABE = True


class TestUnitConversion(unittest.TestCase):
    """
    Test for Unit Conversion
    """

    def setUp(self) -> None:
        """
        Setup the test environment
        """

    def test_import_warning(self) -> None:
        """
        Test for import warning if pint is not installed
        """
        if not IS_PINT_AVAILABE:
            with self.assertRaises(ImportError):
                convert_units("degC", "degF")
        else:
            self.skipTest("Pint installed. Skipping import warning test.")

    def test_example_conversion(self) -> None:
        """
        Test for example conversion when pint is installed
        """
        if IS_PINT_AVAILABE:
            result = convert_units("ug/m^3", "kg/m^3")
            self.assertAlmostEqual(result, 1e-9)
        else:
            self.skipTest("Pint not installed. Skipping conversion test.")



"""
This module contains the builders for the particle representation classes.

Classes:
- MassParticleRepresentationBuilder : Builder class for
    HistogramParticleRepresentation objects.
- PDFParticleRepresentationBuilder : Builder class for
    PDFParticleRepresentation objects.
- DiscreteParticleRepresentationBuilder : Builder class for
    DiscreteParticleRepresentation objects.
"""

from typing import Optional
import logging
from numpy.typing import NDArray
import numpy as np

from particula.abc_builder import (
    BuilderABC,
)
from particula.builder_mixin import (
    BuilderRadiusMixin,
    BuilderConcentrationMixin,
    BuilderMassMixin,
    BuilderDensityMixin,
    BuilderChargeMixin,
    BuilderVolumeMixin,
    BuilderLognormalMixin,
    BuilderParticleResolvedCountMixin,
)
from particula.particles.distribution_strategies import (
    DistributionStrategy,
    RadiiBasedMovingBin,
    ParticleResolvedSpeciatedMass,
)
from particula.particles.activity_strategies import (
    ActivityStrategy,
    ActivityIdealMass,
)
from particula.particles.surface_strategies import (
    SurfaceStrategy,
    SurfaceStrategyVolume,
)
from particula.particles.representation import ParticleRepresentation
from particula.particles.properties.lognormal_size_distribution import (
    lognormal_pdf_distribution,
    lognormal_pmf_distribution,
    lognormal_sample_distribution,
)

logger = logging.getLogger("particula")


# pylint: disable=too-few-public-methods
class BuilderSurfaceStrategyMixin:
    """Mixin class for Builder classes to set surface_strategy.

    Methods:
        - set_surface_strategy : Set the surface_strategy attribute.
    """

    def __init__(self):
        self.surface_strategy = None

    def set_surface_strategy(
        self,
        surface_strategy: SurfaceStrategy,
        surface_strategy_units: Optional[str] = None,
    ):
        """Set the surface strategy of the particle.

        Args:
            - surface_strategy : Surface strategy of the particle.
            - surface_strategy_units : Not used. (for interface consistency)
        """
        if surface_strategy_units is not None:
            logger.warning("Ignoring units for surface strategy parameter.")
        self.surface_strategy = surface_strategy
        return self


# pylint: disable=too-few-public-methods
class BuilderDistributionStrategyMixin:
    """Mixin class for Builder classes to set distribution_strategy.

    Methods:
        - set_distribution_strategy : Set the distribution_strategy attribute.
    """

    def __init__(self):
        self.distribution_strategy = None

    def set_distribution_strategy(
        self,
        distribution_strategy: DistributionStrategy,
        distribution_strategy_units: Optional[str] = None,
    ):
        """Set the distribution strategy of the particle.

        Args:
            - distribution_strategy : Distribution strategy of the particle.
            - distribution_strategy_units : Not used. (for interface)
        """
        if distribution_strategy_units is not None:
            logger.warning(
                "Ignoring units for distribution strategy parameter."
            )
        self.distribution_strategy = distribution_strategy
        return self


# pylint: disable=too-few-public-methods
class BuilderActivityStrategyMixin:
    """Mixin class for Builder classes to set activity_strategy.

    Methods:
        - set_activity_strategy : Set the activity_strategy attribute.
    """

    def __init__(self):
        self.activity_strategy = None

    def set_activity_strategy(
        self,
        activity_strategy: ActivityStrategy,
        activity_strategy_units: Optional[str] = None,
    ):
        """Set the activity strategy of the particle.

        Args:
            - activity_strategy : Activity strategy of the particle.
            - activity_strategy_units : Not used. (for interface consistency)
        """
        if activity_strategy_units is not None:
            logger.warning("Ignoring units for activity strategy parameter.")
        self.activity_strategy = activity_strategy
        return self


class ParticleMassRepresentationBuilder(
    BuilderABC,
    BuilderDistributionStrategyMixin,
    BuilderActivityStrategyMixin,
    BuilderSurfaceStrategyMixin,
    BuilderMassMixin,
    BuilderDensityMixin,
    BuilderConcentrationMixin,
    BuilderChargeMixin,
):  # pylint: disable=too-many-ancestors
    """General ParticleRepresentation objects with mass-based bins.

    Attributes:
        - distribution_strategy : Set the DistributionStrategy.
        - activity_strategy : Set the ActivityStrategy.
        - surface_strategy : Set the SurfaceStrategy.
        - mass : Set the mass of the particles. Default units are 'kg'.
        - density : Set the density of the particles. Default units are
            'kg/m^3'.
        - concentration : Set the concentration of the particles.
            Default units are '1/m^3'.
        - charge : Set the number of charges.
    """

    def __init__(self):
        required_parameters = [
            "distribution_strategy",
            "activity_strategy",
            "surface_strategy",
            "mass",
            "density",
            "concentration",
            "charge",
        ]
        BuilderABC.__init__(self, required_parameters)
        BuilderDistributionStrategyMixin.__init__(self)
        BuilderActivityStrategyMixin.__init__(self)
        BuilderSurfaceStrategyMixin.__init__(self)
        BuilderMassMixin.__init__(self)
        BuilderDensityMixin.__init__(self)
        BuilderConcentrationMixin.__init__(self, default_units="1/m^3")
        BuilderChargeMixin.__init__(self)

    def build(self) -> ParticleRepresentation:
        """Validate and return the ParticleRepresentation object.

        Returns:
            - The validated ParticleRepresentation object.
        """
        self.pre_build_check()
        return ParticleRepresentation(
            strategy=self.distribution_strategy,  # type: ignore
            activity=self.activity_strategy,  # type: ignore
            surface=self.surface_strategy,  # type: ignore
            distribution=self.mass,  # type: ignore
            density=self.density,  # type: ignore
            concentration=self.concentration,  # type: ignore
            charge=self.charge,  # type: ignore
        )


class ParticleRadiusRepresentationBuilder(
    BuilderABC,
    BuilderDistributionStrategyMixin,
    BuilderActivityStrategyMixin,
    BuilderSurfaceStrategyMixin,
    BuilderRadiusMixin,
    BuilderDensityMixin,
    BuilderConcentrationMixin,
    BuilderChargeMixin,
):  # pylint: disable=too-many-ancestors
    """General ParticleRepresentation objects with radius-based bins.

    Attributes:
        - distribution_strategy : Set the DistributionStrategy.
        - activity_strategy : Set the ActivityStrategy.
        - surface_strategy : Set the SurfaceStrategy.
        - radius : Set the radius of the particles. Default units are 'm'.
        - density : Set the density of the particles. Default units are
            'kg/m**3'.
        - concentration : Set the concentration of the particles. Default units
            are '1/m^3'.
        - charge : Set the number of charges.
    """

    def __init__(self):
        required_parameters = [
            "distribution_strategy",
            "activity_strategy",
            "surface_strategy",
            "radius",
            "density",
            "concentration",
            "charge",
        ]
        BuilderABC.__init__(self, required_parameters)
        BuilderDistributionStrategyMixin.__init__(self)
        BuilderActivityStrategyMixin.__init__(self)
        BuilderSurfaceStrategyMixin.__init__(self)
        BuilderRadiusMixin.__init__(self)
        BuilderDensityMixin.__init__(self)
        BuilderConcentrationMixin.__init__(self, default_units="1/m^3")
        BuilderChargeMixin.__init__(self)

    def build(self) -> ParticleRepresentation:
        """Validate and return the ParticleRepresentation object.

        Returns:
            - The validated ParticleRepresentation object.
        """
        self.pre_build_check()
        return ParticleRepresentation(
            strategy=self.distribution_strategy,  # type: ignore
            activity=self.activity_strategy,  # type: ignore
            surface=self.surface_strategy,  # type: ignore
            distribution=self.radius,  # type: ignore
            density=self.density,  # type: ignore
            concentration=self.concentration,  # type: ignore
            charge=self.charge,  # type: ignore
        )


class PresetParticleRadiusBuilder(
    BuilderABC,
    BuilderDistributionStrategyMixin,
    BuilderActivityStrategyMixin,
    BuilderSurfaceStrategyMixin,
    BuilderRadiusMixin,
    BuilderDensityMixin,
    BuilderConcentrationMixin,
    BuilderChargeMixin,
    BuilderLognormalMixin,
):  # pylint: disable=too-many-ancestors
    """General ParticleRepresentation objects with radius-based bins.

    Attributes:
        - mode : Set the mode(s) of the distribution.
            Default is np.array([100e-9, 1e-6]) meters.
        - geometric_standard_deviation : Set the geometric standard
            deviation(s) of the distribution. Default is np.array([1.2, 1.4]).
        - number_concentration : Set the number concentration of the
            distribution. Default is np.array([1e4x1e6, 1e3x1e6])
            particles/m^3.
        - radius_bins : Set the radius bins of the distribution. Default is
            np.logspace(-9, -4, 250), meters.
    """

    def __init__(self):
        required_parameters = [
            "mode",
            "geometric_standard_deviation",
            "number_concentration",
        ]
        BuilderABC.__init__(self, required_parameters)
        BuilderDistributionStrategyMixin.__init__(self)
        BuilderActivityStrategyMixin.__init__(self)
        BuilderSurfaceStrategyMixin.__init__(self)
        BuilderRadiusMixin.__init__(self)
        BuilderDensityMixin.__init__(self)
        BuilderConcentrationMixin.__init__(self, default_units="1/m^3")
        BuilderChargeMixin.__init__(self)
        BuilderLognormalMixin.__init__(self)

        # set defaults
        self.mode = np.array([100e-9, 1e-6])
        self.geometric_standard_deviation = np.array([1.2, 1.4])
        self.number_concentration = np.array([1e4 * 1e6, 1e3 * 1e6])
        self.radius_bins = np.logspace(-9, -4, 250)
        self.set_distribution_strategy(RadiiBasedMovingBin())
        self.set_activity_strategy(ActivityIdealMass())
        self.set_surface_strategy(
            SurfaceStrategyVolume(surface_tension=0.072, density=1000)
        )
        self.set_density(1000, "kg/m^3")
        self.set_charge(0)
        self.distribution_type = "pmf"

    def set_radius_bins(
        self,
        radius_bins: NDArray[np.float64],
        radius_bins_units: str = "m",
    ):
        """Set the radius bins for the distribution

        Arguments:
            - radius_bins : The radius bins for the distribution.
        """
        if np.any(radius_bins < 0):
            message = "The radius bins must be positive."
            logger.error(message)
            raise ValueError(message)
        if radius_bins_units == "m":
            self.radius_bins = radius_bins
            return self
        raise ValueError("Only meters are supported for radius bins.")

    def set_distribution_type(
        self,
        distribution_type: str,
        distribution_type_units: Optional[str] = None,
    ):
        """Set the distribution type for the particle representation.

        Arguments:
            - distribution_type : The type of distribution to use.
        """
        if distribution_type not in ["pdf", "pmf"]:
            message = "The distribution type must be either 'pdf' or 'pmf'."
            logger.error(message)
            raise ValueError(message)
        if distribution_type_units is not None:
            logger.warning("Ignoring units for distribution type parameter.")
        self.distribution_type = distribution_type
        return self

    def build(self) -> ParticleRepresentation:
        """Validate and return the ParticleRepresentation object.

        This will build a distribution of particles with a lognormal size
        distribution, before returning the ParticleRepresentation object.

        Returns:
            - The validated ParticleRepresentation object.
        """
        if self.distribution_type == "pdf":
            number_concentration = lognormal_pdf_distribution(
                x_values=self.radius_bins,
                mode=self.mode,
                geometric_standard_deviation=self.geometric_standard_deviation,
                number_of_particles=self.number_concentration,
            )
        elif self.distribution_type == "pmf":
            number_concentration = lognormal_pmf_distribution(
                x_values=self.radius_bins,
                mode=self.mode,
                geometric_standard_deviation=self.geometric_standard_deviation,
                number_of_particles=self.number_concentration,
            )
        else:
            message = "The distribution type must be either 'pdf' or 'pmf'."
            logger.error(message)
            raise ValueError(message)

        self.pre_build_check()
        return ParticleRepresentation(
            strategy=self.distribution_strategy,  # type: ignore
            activity=self.activity_strategy,  # type: ignore
            surface=self.surface_strategy,  # type: ignore
            distribution=self.radius_bins,
            density=self.density,  # type: ignore
            concentration=number_concentration,  # type: ignore
            charge=self.charge,  # type: ignore
        )


class ResolvedParticleMassRepresentationBuilder(
    BuilderABC,
    BuilderDistributionStrategyMixin,
    BuilderActivityStrategyMixin,
    BuilderSurfaceStrategyMixin,
    BuilderDensityMixin,
    BuilderChargeMixin,
    BuilderVolumeMixin,
    BuilderMassMixin,
):  # pylint: disable=too-many-ancestors
    """
    Builder class for constructing ParticleRepresentation objects with
    resolved masses.

    This class allows you to set various attributes for a particle
    representation, such as distribution strategy, mass, density, charge,
    volume, and more. These attributes are validated and there a no presets.

    Attributes:
        - distribution_strategy : Set the distribution strategy for particles.
        - activity_strategy : Set the activity strategy for the particles.
        - surface_strategy : Set the surface strategy for the particles.
        - mass : Set the particle mass. Defaults to 'kg'.
        - density : Set the particle density. Defaults to 'kg/m^3'.
        - charge : Set the particle charge.
        - volume : Set the particle volume. Defaults to 'm^3'.
    """

    def __init__(self):
        required_parameters = [
            "distribution_strategy",
            "activity_strategy",
            "surface_strategy",
            "mass",
            "density",
            "charge",
            "volume",
        ]
        BuilderABC.__init__(self, required_parameters)
        BuilderDistributionStrategyMixin.__init__(self)
        BuilderActivityStrategyMixin.__init__(self)
        BuilderSurfaceStrategyMixin.__init__(self)
        BuilderMassMixin.__init__(self)
        BuilderDensityMixin.__init__(self)
        BuilderChargeMixin.__init__(self)
        BuilderVolumeMixin.__init__(self)

    def build(self) -> ParticleRepresentation:
        """
        Validate and return a ParticleRepresentation object.

        This method validates all the required attributes and builds a particle
        representation with a lognormal size distribution.

        Returns:
            - ParticleRepresentation : A validated particle representation
                object.
        """
        number_concentration = np.ones_like(  # type: ignore
            self.mass, dtype=np.float64  # type: ignore
        )
        if number_concentration.ndim > 1:
            number_concentration = number_concentration[:, 0]

        self.pre_build_check()
        return ParticleRepresentation(
            strategy=self.distribution_strategy,  # type: ignore
            activity=self.activity_strategy,  # type: ignore
            surface=self.surface_strategy,  # type: ignore
            distribution=self.mass,  # type: ignore
            density=self.density,  # type: ignore
            concentration=number_concentration,  # type: ignore
            charge=self.charge,  # type: ignore
            volume=self.volume,  # type: ignore
        )


class PresetResolvedParticleMassBuilder(
    BuilderABC,
    BuilderDistributionStrategyMixin,
    BuilderActivityStrategyMixin,
    BuilderSurfaceStrategyMixin,
    BuilderDensityMixin,
    BuilderChargeMixin,
    BuilderLognormalMixin,
    BuilderVolumeMixin,
    BuilderParticleResolvedCountMixin,
):  # pylint: disable=too-many-ancestors
    """General ParticleRepresentation objects with particle resolved masses.

    This class has preset values for all the attributes, and allows you to
    override them as needed. This is useful when you want to quickly
    particle representation object with resolved masses.

    Attributes:
        - distribution_strategy : Set the DistributionStrategy.
        - activity_strategy : Set the ActivityStrategy.
        - surface_strategy : Set the SurfaceStrategy.
        - mass : Set the mass of the particles Default
            units are 'kg'.
        - density : Set the density of the particles.
            Default units are 'kg/m^3'.
        - charge : Set the number of charges.
        - mode : Set the mode(s) of the distribution.
            Default is np.array([100e-9, 1e-6]) meters.
        - geometric_standard_deviation : Set the geometric standard
            deviation(s) of the distribution. Default is np.array([1.2, 1.4]).
        - number_concentration : Set the number concentration of the
            distribution. Default is np.array([1e4 1e6, 1e3 1e6])
            particles/m^3.
        - particle_resolved_count : Set the number of resolved particles.
    """

    def __init__(self):
        required_parameters = [
            "mode",
            "geometric_standard_deviation",
            "number_concentration",
            "particle_resolved_count",
            "volume",
        ]
        BuilderABC.__init__(self, required_parameters)
        BuilderDistributionStrategyMixin.__init__(self)
        BuilderActivityStrategyMixin.__init__(self)
        BuilderSurfaceStrategyMixin.__init__(self)
        BuilderDensityMixin.__init__(self)
        BuilderChargeMixin.__init__(self)
        BuilderLognormalMixin.__init__(self)
        BuilderVolumeMixin.__init__(self)
        BuilderParticleResolvedCountMixin.__init__(self)

        # set defaults
        self.mode = np.array([100e-9, 1e-6])
        self.geometric_standard_deviation = np.array([1.2, 1.4])
        self.number_concentration = np.array([1e4 * 1e6, 1e3 * 1e6])
        self.particle_resolved_count = int(10_000)
        self.set_distribution_strategy(ParticleResolvedSpeciatedMass())
        self.set_activity_strategy(ActivityIdealMass())
        self.set_surface_strategy(
            SurfaceStrategyVolume(surface_tension=0.072, density=1000)
        )
        self.set_density(1000, "kg/m^3")
        self.set_charge(0)
        self.set_volume(1)

    def build(self) -> ParticleRepresentation:
        """Validate and return the ParticleRepresentation object.

        This will build a distribution of particles with a lognormal size
        distribution, before returning the ParticleRepresentation object.

        Returns:
            - The validated ParticleRepresentation object.
        """

        resolved_radii = lognormal_sample_distribution(
            mode=self.mode,
            geometric_standard_deviation=self.geometric_standard_deviation,
            number_of_particles=self.number_concentration,
            number_of_samples=self.particle_resolved_count,
        )
        # convert radii to masses
        resolved_masses = np.float64(
            4 / 3 * np.pi * resolved_radii**3 * self.density  # type: ignore
        )
        number_concentration = np.ones_like(resolved_masses, dtype=np.float64)

        self.pre_build_check()
        return ParticleRepresentation(
            strategy=self.distribution_strategy,  # type: ignore
            activity=self.activity_strategy,  # type: ignore
            surface=self.surface_strategy,  # type: ignore
            distribution=resolved_masses,  # type: ignore
            density=self.density,  # type: ignore
            concentration=number_concentration,
            charge=self.charge,  # type: ignore
            volume=self.volume,  # type: ignore
        )
